from app.readers import tsv as reader
from app.actions.prottable import isoquant as preparation
from app.drivers.prottable.base import ProttableAddData
from app.drivers.options import prottable_options


class AddIsobaricQuantDriver(ProttableAddData):
    outsuffix = '_added_isoq.txt'
    command = 'addisoquant'
    commandhelp = ('Add isobaric quantification data from another '
                   'proteintable containing this.')

    def set_options(self):
        super().set_options()
        self.options.update(self.define_options(['quantfile',
                                                 'quantacccolpattern',
                                                 'quantcolpattern'],
                                                prottable_options))

    def initialize_input(self):
        super().initialize_input()
        quantheader = reader.get_tsv_header(self.quantfile)
        self.quantfields = reader.get_cols_in_file(self.quantcolpattern,
                                                   quantheader)
        self.quantacc = reader.get_cols_in_file(self.quantacccolpattern,
                                                quantheader, single_col=True)
        self.quantfeatures = reader.generate_tsv_proteins(self.quantfile,
                                                          quantheader)

    def create_header(self):
        self.header = self.oldheader + self.quantfields

    def set_feature_generator(self):
        self.features = preparation.add_isoquant_data(self.in_proteins,
                                                      self.quantfeatures,
                                                      self.quantacc,
                                                      self.quantfields)



HEADER_PEPTIDE = 'Peptide sequence'
HEADER_NO_PSM = '# PSMs'
HEADER_AREA = 'MS1 area (highest of all PSMs)'
HEADER_QVAL = 'q-value'
HEADER_QVAL_MODELED = 'q-value (linear modeled)'
HEADER_PEP = 'PEP'
HEADER_LINKED_PSMS = 'Files/scans for peptide'
HEADER_PROTEINS = 'Protein(s)'
HEADER_MASTERPROTEINS = 'Master protein(s)'
HEADER_NO_CONTENTPROTEINS = '# proteins in group'
HEADER_COVERAGES = 'Coverage(s)'
HEADER_GENES = 'Gene(s)'
HEADER_ASSOCIATED = 'Associated gene ID(s)'
HEADER_DESCRIPTIONS = 'Description(s)'
HEADER_SVM_SCORE = 'percolator svm-score'



from app.actions.mslookup import proteinquant as lookups
from app.drivers.mslookup import base
from app.drivers.options import mslookup_options


class ProteinQuantLookupDriver(base.LookupDriver):
    """Creates lookup of protein tables that contain quant data"""
    lookuptype = 'prottable'
    command = 'proteins'
    commandhelp = ('Creates lookup of protein quantification '
                   'data in tab separated format. Header should include '
                   'quantification channel names, and if possible the '
                   'number of peptides quantified for each protein in the '
                   'respective channels. Can be used with --genecentric '
                   'Lookup should already include proteins.')

    def __init__(self):
        super().__init__()
        self.infiletype = 'protein table'

    def set_options(self):
        super().set_options()
        self.options.update(self.define_options(['genecentric', 'setnames',
                                                 'quantcolpattern',
                                                 'psmnrcolpattern',
                                                 'precursorquantcolpattern',
                                                 'probcolpattern',
                                                 'fdrcolpattern',
                                                 'pepcolpattern',
                                                 'multifiles', 'proteincol'],
                                                mslookup_options))

    def parse_input(self, **kwargs):
        super().parse_input(**kwargs)
        if self.genecentric in ['genes', 'assoc']:
            # cannot simply do if genecentric: because sometimes it is
            # True e.g. from a child class and then a KeyError arises below
            self.lookuptype = {'genes': 'genetable',
                               'assoc': 'associdtable'}[self.genecentric]
        self.setnames = [x.replace('"', '') for x in self.setnames]
        # FIXME need check to see same poolnames correlate with self.fn len

    def create_lookup(self):
        self.proteincol = self.proteincol - 1
        lookups.create_proteinquant_lookup(self.fn, self.lookup,
                                           self.setnames,
                                           self.proteincol,
                                           self.precursorquantcolpattern,
                                           self.quantcolpattern,
                                           self.psmnrcolpattern,
                                           self.probcolpattern,
                                           self.fdrcolpattern,
                                           self.pepcolpattern)



MZIDTSV_PEP_COL = 9
MZIDTSV_PROT_COL = 10
DB_STORE_CHUNK = 100000

from collections import OrderedDict

from app.readers import tsv as tsvreader
from app.actions.mzidtsv import proteingroup_sorters as sorters


def build_proteingroup_db(pgdb):
    build_master_db(pgdb)
    build_coverage(pgdb)
    build_content_db(pgdb)


def build_master_db(pgdb):
    psm_masters = OrderedDict()
    allmasters = {}
    allpsms = {}
    for psmid, protein in pgdb.get_all_psm_protein_relations():
        try:
            allpsms[psmid].append(protein)
        except KeyError:
            allpsms[psmid] = [protein]
    while len(allpsms) > 0:
        psm_id, proteins = allpsms.popitem()
        pepprotmap = pgdb.get_protpepmap_from_proteins(proteins)
        masters = get_masters(pepprotmap)
        for psm, master in [(p, m) for p, pmasters in masters.items()
                            for m in pmasters]:
            try:
                psm_masters[psm].add(master)
            except KeyError:
                psm_masters[psm] = set([master])
            allmasters[master] = 1
            if psm in allpsms:
                del(allpsms[psm])
    print('Collected {0} masters, {1} PSM-master mappings'.format(
        len(allmasters), len(psm_masters)))
    pgdb.store_masters(allmasters, psm_masters)


def process_pgroup_candidates(candidates, protein_psm_map):
    prepgroup = {}
    for candidate in candidates:
        master, psm_id, prot_id, seq, score, evid, cov = candidate
        prepgroup = add_protein_psm_to_pre_proteingroup(prepgroup, prot_id,
                                                        seq, psm_id, score,
                                                        evid, cov)
    pgroup = filter_proteins_with_missing_psms(prepgroup, protein_psm_map)
    return get_protein_group_content(pgroup, master)


def build_content_db(pgdb):
    protein_psms = {}
    for prot, psm in pgdb.get_protein_psm_records():
        try:
            protein_psms[prot].add(psm)
        except KeyError:
            protein_psms[prot] = set([psm])
    use_evi = pgdb.check_evidence_tables()
    pg_candidates = pgdb.get_protein_group_candidates()
    pre_protein_group = [next(pg_candidates)]
    lastmaster = pre_protein_group[0][0]
    protein_groups, new_masters = [], {}
    for protein_candidate in pg_candidates:
        if protein_candidate[0] != lastmaster:
            pgroup = process_pgroup_candidates(pre_protein_group, protein_psms)
            new_master = sorters.sort_to_get_master(pgroup, use_evi)
            new_masters[new_master['master_id']] = new_master['protein_acc']
            protein_groups.extend(pgroup)
            lastmaster, pre_protein_group = protein_candidate[0], []
        pre_protein_group.append(protein_candidate)
    pgroup = process_pgroup_candidates(pre_protein_group, protein_psms)
    new_master = sorters.sort_to_get_master(pgroup, use_evi)
    new_masters[new_master['master_id']] = new_master['protein_acc']
    protein_groups.extend(pgroup)
    protein_groups = [[pg[2], pg[1], pg[3], pg[4], pg[5]]
                      for pg in protein_groups]
    new_masters = ((acc, mid) for mid, acc in new_masters.items())
    pgdb.update_master_proteins(new_masters)
    pgdb.store_protein_group_content(protein_groups)
    pgdb.index_protein_group_content()


def add_protein_psm_to_pre_proteingroup(prepgmap, protein, pepseq,
                                        psm_id, score, evid, cover):
    protpsm_unit = (psm_id, float(score), evid, cover)
    try:
        prepgmap[protein][pepseq].add(protpsm_unit)
    except KeyError:
        try:
            prepgmap[protein][pepseq] = {protpsm_unit}
        except KeyError:
            prepgmap[protein] = {pepseq: {protpsm_unit}}
    return prepgmap


def filter_proteins_with_missing_psms(proteins, allprotein_psms):
    filtered_protein_map = {}
    for protein, protein_psms in proteins.items():
        protein_masterpsms = {psm[0] for peptide in protein_psms.values()
                              for psm in peptide}
        if allprotein_psms[protein].difference(protein_masterpsms):
            continue
        else:
            filtered_protein_map[protein] = protein_psms
    return filtered_protein_map


def build_coverage(pgdb):
    coverage = {}
    for acc, seq, psm_id, psmseq in pgdb.get_all_proteins_psms_seq():
        try:
            coverage[acc]['seq'] = seq
        except KeyError:
            coverage[acc] = {'seq': seq, 'psms': []}
        coverage[acc]['psms'].append(psmseq)
    pgdb.store_coverage(generate_coverage(coverage))


def get_masters(ppgraph):
    """From a protein-peptide graph dictionary (keys proteins,
    values peptides), return master proteins aka those which
    have no proteins whose peptides are supersets of them.
    If shared master proteins are found, report only the first,
    we will sort the whole proteingroup later anyway. In that
    case, the master reported here may be temporary."""
    masters = {}
    for protein, peps in ppgraph.items():
        ismaster = True
        peps = set(peps)
        multimaster = set()
        for subprotein, subpeps in ppgraph.items():
            if protein == subprotein:
                continue
            if peps.issubset(subpeps):
                if peps.union(subpeps) > peps:
                    ismaster = False
                    break
                elif peps.intersection(subpeps) == peps:
                    multimaster.update({protein, subprotein})
        if not ismaster:
            continue
        elif multimaster:
            premaster = sorted(list(multimaster))[0]
        else:
            premaster = protein
        for pep in peps:
            try:
                masters[pep].add(premaster)
            except KeyError:
                masters[pep] = {premaster}
    return masters


def generate_coverage(seqinfo):
    """From a dict containing protein accessions and sequences/PSM sequences,
    this function returns a generator that calculates coverages for each
    protein and returns the accession and coverage percentage.
    Coverage is done by finding peptides in the protein seq using seq.index
    and marking the range. May be slow."""
    for acc, protinfo in seqinfo.items():
        coverage_aa_indices = set()
        seq = protinfo['seq']
        for psmseq in protinfo['psms']:
            psmseq = tsvreader.strip_modifications(psmseq)
            # FIXME try block is for problems with coverage, see if it is
            # needed
            try:
                start = seq.index(psmseq)
            except:
                print('CANNOT FIND PSM seq {0} in seq {1} '
                      'for acc {2}'.format(psmseq, seq, acc))
            coverage_aa_indices.update(range(start, start + len(psmseq)))
        yield (acc, len(coverage_aa_indices) / len(seq))


def get_protein_group_content(pgmap, master):
    """For each master protein, we generate the protein group proteins
    complete with sequences, psm_ids and scores. Master proteins are included
    in this group.

    Returns a list of [protein, master, pep_hits, psm_hits, protein_score],
    which is ready to enter the DB table.
    """
    # first item (0) is only a placeholder so the lookup.INDEX things get the
    # correct number. Would be nice with a solution, but the INDEXes were
    # originally made for mzidtsv protein group adding.
    pg_content = [[0, master, protein, len(peptides), len([psm for pgpsms in
                                                           peptides.values()
                                                           for psm in pgpsms]),
                   sum([psm[1] for pgpsms in peptides.values()
                        for psm in pgpsms]),  # score
                   next(iter(next(iter(peptides.values()))))[3],  # coverage
                   next(iter(next(iter(peptides.values()))))[2],  # evid level
                   ]
                  for protein, peptides in pgmap.items()]
    return pg_content



from app.actions.mzidtsv import confidencefilters as conffilt
from app.readers import tsv as tsvreader


def generate_psms(fn, header, confkey, conflvl, lower_is_better):
    for psm in tsvreader.generate_tsv_psms(fn, header):
        if conffilt.passes_filter(psm, conflvl, confkey, lower_is_better):
            yield psm



from app.actions.prottable import info as preparation
from app.drivers.prottable.base import ProttableAddData
from app.drivers.options import prottable_options
from app.writers import prottable as writers


class AddProteinInfoDriver(ProttableAddData):
    outsuffix = '_proteindata.txt'
    lookuptype = 'prottable'
    command = 'proteindata'
    commandhelp = ('Add protein data (description, coverage,# PSMs, etc.) '
                   'to a table with protein accessions')

    def set_options(self):
        super().set_options()
        self.options.update(self.define_options(['setname', 'genecentric',
                                                 'lookupfn'],
                                                prottable_options))

    def parse_input(self, **kwargs):
        super().parse_input(**kwargs)
        self.headertypes = ['proteindata']
        self.poolnames = [kwargs.get('setname')]
        if self.genecentric:
            self.lookuptype = {'genes': 'genetable',
                               'assoc': 'associdtable'}[self.genecentric]

    def set_feature_generator(self):
        self.features = preparation.add_protein_data(self.in_proteins,
                                                     self.lookup,
                                                     self.headerfields,
                                                     self.genecentric,
                                                     self.setname
                                                     )

    def write(self):
        outfn = self.create_outfilepath(self.fn, self.outsuffix)
        writers.write_prottable_with_na(self.header, self.features, outfn)



from app.dataformats import peptable as peptabledata
from app.dataformats import prottable as prottabledata


def add_nesvi_protein_probability(proteins, peptides, headerfields, protcol):
    protein_probs = {}
    if not protcol:
        protcol = peptabledata.HEADER_MASTERPROTEINS
    for peptide in peptides:
        protacc = peptide[protcol]
        pep = peptide[peptabledata.HEADER_PEP]
        if ';' in protacc or pep in ['NA', False]:
            continue
        pep = float(pep)
        try:
            protein_probs[protacc] = protein_probs[protacc] * pep
        except KeyError:
            protein_probs[protacc] = pep
    for protein in proteins:
        outprotein = {k: v for k, v in protein.items()}
        try:
            protprob = protein_probs[protein[prottabledata.HEADER_PROTEIN]]
        except KeyError:
            protprob = 'NA'
        outprotein[headerfields['probability']
                   [prottabledata.HEADER_PROBABILITY][None]] = str(protprob)
        yield outprotein



from collections import OrderedDict

from app.dataformats import mzidtsv as mzidtsvdata
from app.readers import tsv as tsvreader
from app.readers import fasta as fastareader
DB_STORE_CHUNK = 100000


def create_psm_lookup(fn, fastafn, mapfn, header, pgdb, unroll=False,
                      specfncol=None, decoy=False,
                      fastadelim=None, genefield=None):
    """Reads PSMs from file, stores them to a database backend in chunked PSMs.
    """
    proteins = store_proteins_descriptions(pgdb, fastafn, fn, mapfn, header,
                                           decoy, fastadelim, genefield)
    mzmlmap = pgdb.get_mzmlfile_map()
    sequences = {}
    for psm in tsvreader.generate_tsv_psms(fn, header):
        seq = tsvreader.get_psm_sequence(psm, unroll)
        sequences[seq] = 1
    pgdb.store_pepseqs(((seq,) for seq in sequences))
    pepseqmap = pgdb.get_peptide_seq_map()
    psms = []
    for row, psm in enumerate(tsvreader.generate_tsv_psms(fn, header)):
        specfn, psm_id, scan, seq, score = tsvreader.get_psm(psm, unroll,
                                                             specfncol)
        if len(psms) % DB_STORE_CHUNK == 0:
            pgdb.store_psms(psms)
            psms = []
        psms.append({'rownr': row,
                     'psm_id': psm_id,
                     'seq': pepseqmap[seq],
                     'score': score,
                     'specfn': mzmlmap[specfn],
                     'scannr': scan,
                     'spec_id': '{}_{}'.format(mzmlmap[specfn], scan),
                     })
    pgdb.store_psms(psms)
    pgdb.index_psms()
    store_psm_protein_relations(fn, header, pgdb, proteins)


def get_protein_gene_map(mapfn, proteins, decoy):
    gpmap = {}
    for protein, gene, symbol, desc in fastareader.get_proteins_genes(mapfn):
        if protein in proteins:
            protein = proteins[protein][0]
            if decoy:
                symbol = '{}{}'.format(mzidtsvdata.DECOY_PREFIX, symbol)
                gene = '{}{}'.format(mzidtsvdata.DECOY_PREFIX, gene)
                desc = None
            gpmap[protein] = {'gene': gene, 'symbol': symbol, 'desc': desc}
    return gpmap


def store_proteins_descriptions(pgdb, fastafn, tsvfn, mapfn, header, decoy,
                                fastadelim, genefield):
    if not fastafn:
        proteins = {}
        for psm in tsvreader.generate_tsv_psms(tsvfn, header):
            proteins.update({x: 1 for x in
                             tsvreader.get_proteins_from_psm(psm)})
        proteins = [(protein,) for protein in proteins.keys()]
        pgdb.store_proteins(proteins)
    else:
        proteins, sequences, evidences = fastareader.get_proteins_for_db(
            fastafn)
        proteins = [x for x in proteins]
        pgdb.store_proteins(proteins, evidences, sequences)
        if not mapfn:
            associations = fastareader.get_proteins_genes(fastafn, fastadelim,
                                                          genefield)
            genes, descriptions = [], []
            for assoc in associations:
                genes.append((assoc[1], assoc[0]))
                descriptions.append((assoc[0], assoc[3]))
            pgdb.store_descriptions(descriptions)
            pgdb.store_genes(genes)
    if mapfn:
        proteins_with_versions = {}
        for protein in proteins:
            proteins_with_versions[protein[0].split('.')[0]] = protein
        if decoy:
            mod = fastareader.get_decoy_mod_string(proteins[0][0])
            proteins_with_versions = {k.replace(mod, ''): v for k, v
                                      in proteins_with_versions.items()}
        gpmap = get_protein_gene_map(mapfn, proteins_with_versions, decoy)
        pgdb.store_gene_and_associated_id(gpmap)
    return set([x[0] for x in proteins])


def store_psm_protein_relations(fn, header, pgdb, proteins):
    """Reads PSMs from file, extracts their proteins and peptides and passes
    them to a database backend in chunks.
    """
    # TODO do we need an OrderedDict or is regular dict enough?
    # Sorting for psm_id useful?
    allpsms = OrderedDict()
    last_id, psmids_to_store = None, set()
    store_soon = False
    for psm in tsvreader.generate_tsv_psms(fn, header):
        psm_id, prots = tsvreader.get_pepproteins(psm)
        prots = [x for x in prots if x in proteins]
        try:
            # In case the PSMs are presented unrolled
            allpsms[psm_id].extend(prots)
        except KeyError:
            allpsms[psm_id] = prots
        if len(psmids_to_store) % DB_STORE_CHUNK == 0:
            store_soon = True
        if store_soon and last_id != psm_id:
            pgdb.store_peptides_proteins(allpsms, psmids_to_store)
            store_soon = False
            psmids_to_store = set()
        psmids_to_store.add(psm_id)
        last_id = psm_id
    if len(psmids_to_store) > 0:
        pgdb.store_peptides_proteins(allpsms, psmids_to_store)
    pgdb.index_protein_peptides()
    return allpsms



from app.lookups.sqlite.biosets import BioSetDB


class SpectraDB(BioSetDB):
    def add_tables(self):
        super().add_tables()
        self.create_tables(['mzml'])

    def store_mzmls(self, spectra):
        self.store_many(
            'INSERT INTO mzml(spectra_id, mzmlfile_id, scan_nr, charge, mz, '
            'retention_time, ion_injection_time) '
            'VALUES (?, ?, ?, ?, ?, ?, ?)', spectra)

    def index_mzml(self):
        self.index_column('spectra_id_index', 'mzml', 'spectra_id')
        self.index_column('mzmlfnid_mzml_index', 'mzml', 'mzmlfile_id')
        self.index_column('scan_index', 'mzml', 'scan_nr')
        self.index_column('specrt_index', 'mzml', 'retention_time')
        self.index_column('specmz_index', 'mzml', 'mz')

    def get_exp_spectra_data_rows(self):
        cursor = self.get_cursor()
        return cursor.execute('SELECT pr.rownr, bs.set_name, sp.retention_time, '
                              'sp.ion_injection_time '
                              'FROM psmrows AS pr '
                              'JOIN psms AS p USING(psm_id) '
                              'JOIN mzml AS sp USING(spectra_id) '
                              'JOIN mzmlfiles as mf USING(mzmlfile_id) '
                              'JOIN biosets AS bs USING(set_id) '
                              'ORDER BY pr.rownr')



from app.dataformats import prottable as prottabledata
from app.dataformats import mzidtsv as mzidtsvdata


def generate_top_psms(psms, protcol):
    """Fed with a psms generator, this returns the 3 PSMs with
    the highest precursor intensities (or areas, or whatever is
    given in the HEADER_PRECURSOR_QUANT"""
    top_ms1_psms = {}
    for psm in psms:
        protacc = psm[protcol]
        precursor_amount = psm[mzidtsvdata.HEADER_PRECURSOR_QUANT]
        if ';' in protacc or precursor_amount == 'NA':
            continue
        precursor_amount = float(precursor_amount)
        psm_seq = psm[mzidtsvdata.HEADER_PEPTIDE]
        try:
            peptide_area = top_ms1_psms[protacc][psm_seq]
        except KeyError:
            try:
                top_ms1_psms[protacc][psm_seq] = precursor_amount
            except KeyError:
                top_ms1_psms[protacc] = {psm_seq: precursor_amount}
        else:
            if precursor_amount > peptide_area:
                top_ms1_psms[protacc][psm_seq] = precursor_amount
    return top_ms1_psms


def calculate_protein_precursor_quant(top_ms1_psms, prot_acc):
    try:
        amounts = top_ms1_psms[prot_acc].values()
    except KeyError:
        return 'NA'
    else:
        amounts = sorted([x for x in amounts if x > 0], reverse=True)[:3]
        return sum(amounts) / len(amounts)


def add_ms1_quant_from_top3_mzidtsv(proteins, psms, headerfields, protcol):
    """Collects PSMs with the highes precursor quant values,
    adds sum of the top 3 of these to a protein table"""
    if not protcol:
        protcol = mzidtsvdata.HEADER_MASTER_PROT
    top_ms1_psms = generate_top_psms(psms, protcol)
    for protein in proteins:
        prot_acc = protein[prottabledata.HEADER_PROTEIN]
        prec_area = calculate_protein_precursor_quant(top_ms1_psms, prot_acc)
        outprotein = {k: v for k, v in protein.items()}
        outprotein[headerfields['precursorquant'][
            prottabledata.HEADER_AREA][None]] = str(prec_area)
        yield outprotein



from app.actions.mzidtsv import prot2gene as actions
from app.drivers.mzidtsv import MzidTSVDriver
from app.drivers.options import mzidtsv_options


class TSVGeneFromProteinDriver(MzidTSVDriver):
    outsuffix = '_genes.txt'
    lookuptype = 'psm'
    command = 'genes'
    commandhelp = ('Add column to mzidtsv with gene names or symbols, '
                   'which are stored in a lookup specified with --dbfile')

    def set_options(self):
        super().set_options()
        self.options.update(self.define_options(['lookupfn'], mzidtsv_options))

    def get_psms(self):
        """Creates iterator to write to new tsv. Contains input tsv
        lines plus quant data for these."""
        self.header = actions.create_header(self.oldheader)
        self.psms = actions.add_genes_to_psm_table(self.fn, self.oldheader,
                                                   self.lookup)



from app.readers import tsv as reader


def base_add_isoquant_data(features, quantfeatures, acc_col, quantacc_col,
                           quantfields):
    """Generic function that takes a peptide or protein table and adds
    quant data from ANOTHER such table."""
    quant_map = get_quantmap(quantfeatures, quantacc_col, quantfields)
    for feature in features:
        feat_acc = feature[acc_col]
        outfeat = {k: v for k, v in feature.items()}
        try:
            outfeat.update(quant_map[feat_acc])
        except KeyError:
            outfeat.update({field: 'NA' for field in quantfields})
        yield outfeat


def get_quantmap(features, acc_col, quantfields):
    """Runs through proteins that are in a quanted protein table, extracts
    and maps their information based on the quantfields list input.
    Map is a dict with protein_accessions as keys."""
    qmap = {}
    for feature in features:
        feat_acc = feature.pop(acc_col)
        qmap[feat_acc] = {qf: feature[qf] for qf in quantfields}
    return qmap




from app.dataformats import mzidtsv as psmtsvdata


def add_peptide(allpeps, psm, key, score, fncol=False, new=False):
    peptide = {'score': score,
               'line': psm,
               'psms': []
               }
    if fncol:
        if not new:
            peptide['psms'] = allpeps[key]['psms']
    allpeps[key] = peptide


def add_psm_link(peptide, psm, fncol):
    peptide['psms'].append('{0}_{1}'.format(psm[fncol],
                                            psm[psmtsvdata.HEADER_SCANNR]))


def evaluate_peptide(peptides, psm, key, higherbetter, scorecol, fncol=False):
    try:
        score = float(psm[scorecol])
    except ValueError:
        # If score is NA or similar, dont use this PSM
        return peptides
    try:
        existing_score = peptides[key]['score']
    except KeyError:
        add_peptide(peptides, psm, key, score, fncol, new=True)
    else:
        if higherbetter and score > existing_score:
            add_peptide(peptides, psm, key, score, fncol)
        elif not higherbetter and score < existing_score:
            add_peptide(peptides, psm, key, score, fncol)
    if fncol:
        add_psm_link(peptides[key], psm, fncol)
    return peptides



from app.actions.mslookup import proteinquant as lookups
from app.drivers.mslookup.proteinquant import ProteinQuantLookupDriver
from app.drivers.options import mslookup_options, peptable_options


class PeptideQuantLookupDriver(ProteinQuantLookupDriver):
    """Creates lookup of peptide tables that contain quant data"""
    lookuptype = 'peptidetable'
    command = 'peptides'

    def __init__(self):
        super().__init__()
        self.infiletype = 'peptide table'

    def set_options(self):
        super().set_options()
        self.options.update(self.define_options(['peptidecol', 'setnames'],
                                                mslookup_options))
        self.options.update(self.define_options(['genecentric'],
                                                peptable_options))
        self.options.pop('--probcolpattern')
        self.options.pop('--protcol')

    def parse_input(self, **kwargs):
        super().parse_input(**kwargs)
        if self.genecentric:
            self.lookuptype = 'peptidegenecentrictable'

    def create_lookup(self):
        self.peptidecol = self.peptidecol - 1
        lookups.create_peptidequant_lookup(self.fn, self.lookup,
                                           self.setnames,
                                           self.peptidecol,
                                           self.precursorquantcolpattern,
                                           self.quantcolpattern,
                                           self.psmnrcolpattern,
                                           self.fdrcolpattern,
                                           self.pepcolpattern)



from app.actions.mzidtsv import filter_confidence as prep
from app.drivers.mzidtsv import MzidTSVDriver
from app.drivers.options import mzidtsv_options
from app.readers import tsv as tsvreader


class ConfidenceFilterDriver(MzidTSVDriver):
    outsuffix = '_filtconf.txt'
    command = 'conffilt'
    commandhelp = 'Filters PSMs by their confidence level. '

    def set_options(self):
        super().set_options()
        options = self.define_options(['confcol', 'confpattern', 'conflvl',
                                       'conftype', 'unroll'], mzidtsv_options)
        self.options.update(options)

    def parse_input(self, **kwargs):
        super().parse_input(**kwargs)
        self.lowerbetter = self.conftype == 'lower'

    def get_psms(self):
        self.header = self.oldheader[:]
        if self.confpattern:
            confkey = tsvreader.get_cols_in_file(self.confpattern,
                                                 self.header, True)
        elif self.confcol:
            confkey = self.header[int(self.confcol) - 1]
        else:
            raise RuntimeError('Must define either --confcol or '
                               '--confcolpattern')
        self.psms = prep.generate_psms(self.fn,
                                       self.oldheader,
                                       confkey,
                                       self.conflvl,
                                       self.lowerbetter)



from lxml import etree
from app.readers import xml as basereader


def get_percolator_static_xml(fn, ns):
    root = basereader.get_root_el(fn)
    process = etree.iterparse(fn, tag='{%s}process_info' % ns['xmlns'],
                              events=('start',))
    root.append(next(process)[1])
    return root


# Stringified element generators interfaces
def generate_psms_multiple_fractions_strings(input_files, ns):
    return basereader.generate_tags_multiple_files_strings(input_files,
                                                           ns, 'psm',
                                                           ['peptide',
                                                            'protein'])


def generate_peptides_multiple_fractions_strings(input_files, ns):
    return basereader.generate_tags_multiple_files_strings(input_files,
                                                           ns, 'peptide',
                                                           ['psm', 'protein'])


# Element generators interfaces
def generate_psms_multiple_fractions(input_files, ns):
    return basereader.generate_tags_multiple_files(input_files, 'psm',
                                                   ['peptide', 'protein'], ns)


def generate_peptides_multiple_fractions(input_files, ns):
    return basereader.generate_tags_multiple_files(input_files, 'peptide',
                                                   ['psm', 'protein'], ns)


def generate_psms(fn, ns):
    return basereader.generate_xmltags(fn, 'psm', ['peptide', 'protein'], ns)


def generate_peptides(fn, ns):
    return basereader.generate_xmltags(fn, 'peptide', ['psm', 'protein'], ns)


def get_peptide_seq(peptide, ns):
    return peptide.attrib['{%s}peptide_id' % ns['xmlns']]


def get_psm_seq(psm, ns):
    return psm.find('{%s}peptide_seq' % ns['xmlns']).attrib['seq']


def get_psm_ids_from_peptide(peptide, ns):
    return peptide.xpath('xmlns:psm_ids', namespaces=ns)[0]



#!/usr/bin/python3

import sys

from app.drivers.pycolator import (splitmerge, filters, stats, qvality)
from app.drivers import startup


def main():
    drivers = [splitmerge.SplitTDDriver(),
               splitmerge.SplitProteinDriver(),
               splitmerge.MergeDriver(),
               filters.FilterUniquePeptides(),
               filters.FilterPeptideLength(),
               filters.FilterWholeProteinSequence(),
               filters.FilterPeptideSequence(),
               qvality.QvalityDriver(),
               stats.ReassignmentDriver(),
               ]
    startup.start_msstitch(drivers, sys.argv)



from app.dataformats import prottable as prottabledata
from app.actions.prottable import info as pinfo
from app.actions.mergetable import (simple_val_fetch, fill_mergefeature,
                                    get_isobaric_quant)

from app.actions.proteindata import create_featuredata_map


def build_proteintable(pqdb, headerfields, mergecutoff, isobaric=False,
                       precursor=False, probability=False, fdr=False,
                       pep=False, genecentric=False):
    """Fetches proteins and quants from joined lookup table, loops through
    them and when all of a protein's quants have been collected, yields the
    protein quant information."""
    pdmap = create_featuredata_map(pqdb, genecentric=genecentric,
                                   psm_fill_fun=pinfo.add_psms_to_proteindata,
                                   pgene_fill_fun=pinfo.add_protgene_to_protdata,
                                   count_fun=pinfo.count_peps_psms,
                                   get_uniques=True)
    empty_return = lambda x, y, z: {}
    iso_fun = {True: get_isobaric_quant, False: empty_return}[isobaric]
    ms1_fun = {True: get_precursor_quant, False: empty_return}[precursor]
    prob_fun = {True: get_prot_probability,
                False: empty_return}[probability]
    fdr_fun = {True: get_prot_fdr,
               False: empty_return}[fdr]
    pep_fun = {True: get_prot_pep,
               False: empty_return}[pep]
    pdata_fun = {True: get_protein_data_genecentric,
                 False: get_protein_data}[genecentric is not False]
    protein_sql, sqlfieldmap = pqdb.prepare_mergetable_sql(precursor, isobaric,
                                                           probability, fdr,
                                                           pep)
    accession_field = prottabledata.ACCESSIONS[genecentric]
    proteins = pqdb.get_merged_features(protein_sql)
    protein = next(proteins)
    outprotein = {accession_field: protein[sqlfieldmap['p_acc']]}
    check_prot = {k: v for k, v in outprotein.items()}
    if not mergecutoff or protein_pool_fdr_cutoff(protein, sqlfieldmap,
                                                  mergecutoff):
        fill_mergefeature(outprotein, iso_fun, ms1_fun, prob_fun, fdr_fun,
                          pep_fun, pdata_fun, protein, sqlfieldmap,
                          headerfields, pdmap, accession_field)
    for protein in proteins:
        if mergecutoff and not protein_pool_fdr_cutoff(protein, sqlfieldmap,
                                                       mergecutoff):
            continue
        p_acc = protein[sqlfieldmap['p_acc']]
        if p_acc != outprotein[accession_field]:
            # check if protein has been filled, otherwise do not output
            # sometimes proteins have NA in all fields
            if outprotein != check_prot:
                yield outprotein
            outprotein = {accession_field: p_acc}
            check_prot = {k: v for k, v in outprotein.items()}
        fill_mergefeature(outprotein, iso_fun, ms1_fun, prob_fun, fdr_fun,
                          pep_fun, pdata_fun, protein, sqlfieldmap,
                          headerfields, pdmap, accession_field)
    if outprotein != check_prot:
        yield outprotein


def protein_pool_fdr_cutoff(sql_entry, sqlmap, fdrcutoff):
    fdr = sql_entry[sqlmap['fdr_val']]
    warning = False
    if fdr is False:
        warning = True
    try:
        fdr = float(fdr)
    except (TypeError, ValueError):
        warning = True
    if warning:
        print('WARNING, filtering merge table on FDR but protein {} in '
              'set {} has FDR value "{}" in '
              'lookup.'.format(sql_entry[sqlmap['p_acc']],
                               sql_entry[sqlmap['set_name']], fdr))
        return False
    return fdr < fdrcutoff


def get_protein_data_genecentric(outprotein, proteindata_map, headerfields, accfield):
    p_acc = outprotein[accfield]
    return pinfo.get_protein_data_genecentric(proteindata_map, p_acc,
                                              headerfields)


def get_protein_data(outprotein, proteindata_map, headerfields, accfield):
    p_acc = outprotein[accfield]
    return pinfo.get_protein_data_pgrouped(proteindata_map, p_acc,
                                           headerfields)


def get_precursor_quant(protein, sqlmap, headerfields):
    return simple_val_fetch(protein, sqlmap,
                            headerfields['precursorquant'][
                                prottabledata.HEADER_AREA], 'preq_val')


def get_prot_probability(protein, sqlmap, headerfields):
    return simple_val_fetch(protein, sqlmap,
                            headerfields['probability'][
                                prottabledata.HEADER_PROBABILITY], 'prob_val')


def get_prot_fdr(protein, sqlmap, headerfields):
    return simple_val_fetch(protein, sqlmap,
                            headerfields['proteinfdr'][
                                prottabledata.HEADER_QVAL], 'fdr_val')


def get_prot_pep(protein, sqlmap, headerfields):
    return simple_val_fetch(protein, sqlmap,
                            headerfields['proteinpep'][
                                prottabledata.HEADER_PEP], 'pep_val')



"""Reader methods for mzIdentML, tsv as generated by MSGF+"""

from app.readers import xml as basereader
from app.readers import ml

P_SVM = ('percolator:score', 'percolator svm-score')
P_PSMP = ('percolator:psm_p_value', 'PSM p-value')
P_PSMQ = ('percolator:psm_q_value', 'PSM q-value')
P_PSMPEP = ('percolator:psm_pep', 'PSM-PEP')
P_PEPTIDEQ = ('percolator:peptide_q_value', 'peptide q-value')
P_PEPTIDEPEP = ('percolator:peptide_pep', 'peptide PEP')

PERCO_ORDER = [P_SVM, P_PSMP, P_PSMQ, P_PSMPEP, P_PEPTIDEQ, P_PEPTIDEPEP]
PERCO_HEADERMAP = {x[0]: x[1] for x in PERCO_ORDER}


def get_mzid_namespace(mzidfile):
    return basereader.get_namespace_from_top(mzidfile, None)


def mzid_spec_result_generator(mzidfile, namespace):
    return basereader.generate_tags_multiple_files(
        [mzidfile],
        'SpectrumIdentificationResult',
        ['cvList',
         'AnalysisSoftwareList',
         'SequenceCollection',
         'AnalysisProtocolCollection',
         'AnalysisCollection',
         ],
        namespace)


def generate_mzid_peptides(mzidfile, namespace):
    return basereader.generate_tags_multiple_files(
        [mzidfile],
        'Peptide',
        ['cvList',
         'AnalysisSoftwareList',
         'DataCollection',
         'AnalysisProtocolCollection',
         'AnalysisCollection',
         ],
        namespace)


def get_mzid_peptidedata(peptide, xmlns):
    pep_id = peptide.attrib['id']
    sequence = peptide.find('{}PeptideSequence'.format(xmlns)).text
    mods = {}
    for mod in peptide.findall('{}Modification'.format(xmlns)):
        modweight = round(float(mod.attrib['monoisotopicMassDelta']), 3)
        if modweight > 0:
            modweight = '+{}'.format(modweight)
        else:
            modweight = str(modweight)
        location = int(mod.attrib['location'])
        try:
            mods[location] += modweight
        except KeyError:
            mods[location] = modweight
    outseq = []
    for pos, aa in enumerate(sequence):
        if pos in mods:
            outseq.append('{}'.format(mods[pos]))
        outseq.append(aa)
    if pos + 1 in mods:
            outseq.append('{}'.format(mods[pos + 1]))
    return pep_id, ''.join(outseq)


def generate_mzid_spec_id_items(mzidfile, namespace, xmlns, specfn_idmap):
    specid_tag = '{0}SpectrumIdentificationItem'.format(xmlns)
    for specresult in mzid_spec_result_generator(mzidfile, namespace):
        scan = get_specresult_scan_nr(specresult)
        mzmlid = get_specresult_mzml_id(specresult)
        mzmlfn = specfn_idmap[mzmlid]
        for spec_id_item in specresult.findall(specid_tag):
            pep_id = spec_id_item.attrib['peptide_ref']
            yield (scan, mzmlfn, pep_id, spec_id_item)


def mzid_specdata_generator(mzidfile, namespace):
    return basereader.generate_tags_multiple_files(
        [mzidfile],
        'SpectraData',
        ['cvList',
         'AnalysisSoftwareList',
         'SequenceCollection',
         'AnalysisProtocolCollection',
         'AnalysisCollection',
         ],
        namespace)


def get_mzid_specfile_ids(mzidfn, namespace):
    """Returns mzid spectra data filenames and their IDs used in the
    mzIdentML file as a dict. Keys == IDs, values == fns"""
    sid_fn = {}
    for specdata in mzid_specdata_generator(mzidfn, namespace):
        sid_fn[specdata.attrib['id']] = specdata.attrib['name']
    return sid_fn


def get_specresult_scan_nr(result):
    """Returns scan nr of an mzIdentML PSM as a str. The PSM is given
    as a SpectrumIdentificationResult element."""
    return ml.get_scan_nr(result, 'spectrumID')


def get_specresult_mzml_id(specresult):
    return specresult.attrib['spectraData_ref']


def get_specidentitem_percolator_data(item, xmlns):
    """Loop through SpecIdentificationItem children. Find
    percolator data by matching to a dict lookup. Return a
    dict containing percolator data"""
    percomap = {'{0}userParam'.format(xmlns): PERCO_HEADERMAP, }
    percodata = {}
    for child in item:
        try:
            percoscore = percomap[child.tag][child.attrib['name']]
        except KeyError:
            continue
        else:
            percodata[percoscore] = child.attrib['value']
    outkeys = [y for x in list(percomap.values()) for y in list(x.values())]
    for key in outkeys:
        try:
            percodata[key]
        except KeyError:
            percodata[key] = 'NA'
    return percodata



#!/usr/bin/python3

import sys
from app.drivers import startup
from app.drivers.peptable import (psmtopeptable, merge, model_qvals, isoquant)


def main():
    drivers = [psmtopeptable.MzidTSVPeptableDriver(),
               isoquant.AddIsobaricQuantDriver(),
               model_qvals.ModelQValuesDriver(),
               merge.BuildPeptideTableDriver(),
               ]
    startup.start_msstitch(drivers, sys.argv)



#!/usr/bin/python3

import sys
from app.drivers.mslookup import (spectra, quant, proteingroups, biosets,
                                  proteinquant, pepquant, psms, seqspace)
from app.drivers import startup


def main():
    drivers = [biosets.BioSetLookupDriver(),
               spectra.SpectraLookupDriver(),
               psms.PSMLookupDriver(),
               quant.IsobaricQuantLookupDriver(),
               quant.PrecursorQuantLookupDriver(),
               proteingroups.ProteinGroupLookupDriver(),
               pepquant.PeptideQuantLookupDriver(),
               proteinquant.ProteinQuantLookupDriver(),
               seqspace.SeqspaceLookupDriver(),
               seqspace.WholeProteinSeqspaceLookupDriver(),
               ]
    startup.start_msstitch(drivers, sys.argv)



import re
import os
import itertools
from app.dataformats import mzidtsv as mzidtsvdata
from app.dataformats import prottable as prottabledata


def get_tsv_header(tsvfn):
    with open(tsvfn) as fp:
        return next(fp).strip().split('\t')


def generate_tsv_lines_multifile(fns, header):
    return itertools.chain.from_iterable([generate_tsv_psms(fn, header)
                                          for fn in fns])


def generate_tsv_proteins(fn, header):
    return generate_split_tsv_lines(fn, header)


def generate_tsv_pep_protein_quants(fns):
    """Unlike generate_tsv_lines_multifile, this generates tsv lines
    from multiple files that may have different headers. Yields
    fn, header as well as quant data for each protein quant"""
    for fn in fns:
        header = get_tsv_header(fn)
        for pquant in generate_split_tsv_lines(fn, header):
            yield os.path.basename(fn), header, pquant


def generate_kronik_feats(fn):
    """Generates features from a Kronik output file"""
    header = get_tsv_header(fn)
    return generate_split_tsv_lines(fn, header)


def mzmlfn_kronikfeature_generator(mzmlfns, kronikfns):
    """Generates tuples of spectra filename and corresponding output
    features from kronik"""
    for mzmlfn, kronikfn in zip(mzmlfns, kronikfns):
        for quant_el in generate_kronik_feats(kronikfn):
            yield os.path.basename(mzmlfn), quant_el


def generate_tsv_psms(fn, header):
    return generate_split_tsv_lines(fn, header)


def generate_tsv_peptides(fn):
    header = get_tsv_header(fn)
    return generate_split_tsv_lines(fn, header)


def generate_split_tsv_lines(fn, header):
    """Returns dicts with header-keys and psm statistic values"""
    for line in generate_tsv_psms_line(fn):
        yield {x: y for (x, y) in zip(header, line.strip().split('\t'))}


def generate_tsv_psms_line(fn):
    with open(fn) as fp:
        next(fp)  # skip header
        for line in fp:
            yield line


def get_psm_id(line):
    return '{0}_{1}_{2}'.format(line[mzidtsvdata.HEADER_SPECFILE],
                                line[mzidtsvdata.HEADER_SCANNR],
                                line[mzidtsvdata.HEADER_PEPTIDE])


def get_proteins_from_psm(line):
    """From a line, return list of proteins reported by Mzid2TSV. When unrolled
    lines are given, this returns the single protein from the line."""
    proteins = line[mzidtsvdata.HEADER_PROTEIN].split(';')
    outproteins = []
    for protein in proteins:
        prepost_protein = re.sub('\(pre=.*post=.*\)', '', protein).strip()
        outproteins.append(prepost_protein)
    return outproteins


def get_psm(line, unroll=False, specfncol=None):
    """Returns from a PSM line peptide sequence,
    and other information about the PSM.
    Return values:
        specfn          -   str
        psm_id 	 	-   str
        scan            -   str
        peptideseq      -   str
        score		-   str
    """
    if specfncol is None:
        specfncol = mzidtsvdata.HEADER_SPECFILE
    specfn = line[specfncol]
    psm_id = get_psm_id(line)
    scan = line[mzidtsvdata.HEADER_SCANNR]
    peptideseq = get_psm_sequence(line, unroll)
    score = line[mzidtsvdata.HEADER_MSGFSCORE]
    return specfn, psm_id, scan, peptideseq, score


def get_psm_sequence(line, unroll=False):
    peptideseq = line[mzidtsvdata.HEADER_PEPTIDE]
    if unroll and '.' in peptideseq:
        peptideseq = peptideseq.split('.')[1]
    return peptideseq


def get_pepproteins(line):
    """Returns from a PSM line peptide sequence,
    and other information about the PSM.
    Return values:
        psm_id          -   str
        proteins        -   list of str
    """
    psm_id = get_psm_id(line)
    proteins = get_proteins_from_psm(line)
    return psm_id, proteins


def strip_modifications(seq):
    return re.sub('[-+.\d]', '', seq)


def get_cols_in_file(pattern, header, single_col=False):
    if pattern is None:
        return False
    cols_found = get_columns_by_pattern(header, pattern)
    if single_col:
        cols_found = cols_found[0]
    return cols_found


def get_columns_by_pattern(header, pattern):
    columns = []
    for field in header:
        if re.search(pattern, field) is not None:
            columns.append(field)
    if not columns:
        raise RuntimeError('Could not find fieldname in header with '
                           'pattern: {}'.format(pattern))
    return columns



from app.dataformats import prottable as prottabledata
from app.actions.proteindata import (add_psms_to_proteindata,
                                     add_protgene_to_protdata,
                                     create_featuredata_map)


def count_peps_psms(proteindata, p_acc, pool):
    data = proteindata[p_acc]['pools'][pool]
    proteindata[p_acc]['pools'][pool]['psms'] = len(data['psms'])
    proteindata[p_acc]['pools'][pool]['peptides'] = len(data['peptides'])


def add_protein_data(proteins, pgdb, headerfields, genecentric=False,
                     pool_to_output=False):
    """First creates a map with all master proteins with data,
    then outputs protein data dicts for rows of a tsv. If a pool
    is given then only output for that pool will be shown in the
    protein table."""
    proteindata = create_featuredata_map(pgdb, genecentric=genecentric,
                                         psm_fill_fun=add_psms_to_proteindata,
                                         pgene_fill_fun=add_protgene_to_protdata,
                                         count_fun=count_peps_psms,
                                         pool_to_output=pool_to_output,
                                         get_uniques=True)
    dataget_fun = {True: get_protein_data_genecentric,
                   False: get_protein_data_pgrouped}[genecentric is not False]
    firstfield = prottabledata.ACCESSIONS[genecentric]
    for protein in proteins:
        outprotein = {k: v for k, v in protein.items()}
        outprotein[firstfield] = outprotein.pop(prottabledata.HEADER_PROTEIN)
        protein_acc = protein[prottabledata.HEADER_PROTEIN]
        outprotein.update(dataget_fun(proteindata, protein_acc, headerfields))
        outprotein = {k: str(v) for k, v in outprotein.items()}
        yield outprotein


def get_protein_data_genecentric(proteindata, p_acc, headerfields):
    return get_protein_data_base(proteindata, p_acc, headerfields)


def get_protein_data_pgrouped(proteindata, p_acc, headerfields):
    """Parses protein data for a certain protein into tsv output
    dictionary"""
    report = get_protein_data_base(proteindata, p_acc, headerfields)
    return get_cov_protnumbers(proteindata, p_acc, report)


def get_headerfieldtext(headerfields, hfieldtype, pool):
    try:
        text = headerfields['proteindata'][hfieldtype][pool]
    except KeyError:
        text = headerfields['proteindata'][hfieldtype][None]
    return text


def get_protein_data_base(proteindata, p_acc, headerfields):
    hfields = [prottabledata.HEADER_NO_UNIPEP,
               prottabledata.HEADER_NO_PEPTIDE,
               prottabledata.HEADER_NO_PSM,
               ]
    outdict = {prottabledata.HEADER_DESCRIPTION: proteindata[p_acc]['desc']}
    for pool, pdata in proteindata[p_acc]['pools'].items():
        pool_values = [pdata['unipeps'], pdata['peptides'], pdata['psms']]
        outdict.update({get_headerfieldtext(headerfields, hfield, pool): val
                        for (hfield, val) in zip(hfields, pool_values)})
    for field, pdfield in zip([prottabledata.HEADER_GENEID,
                               prottabledata.HEADER_GENENAME,
                               prottabledata.HEADER_PROTEINS],
                              ['gene', 'aid', 'protein_ids']):
        try:
            outdict[field] = ';'.join(proteindata[p_acc][pdfield])
        except TypeError:
            pass
    return outdict


def get_cov_protnumbers(proteindata, p_acc, report):
    try:
        report[prottabledata.HEADER_COVERAGE] = proteindata[p_acc]['cov']
    except KeyError:
        pass
    if 'pgcontent' in proteindata[p_acc]:
        report.update({prottabledata.HEADER_CONTENTPROT:
                       ','.join(proteindata[p_acc]['pgcontent']),
                       prottabledata.HEADER_NO_PROTEIN:
                       len(proteindata[p_acc]['pgcontent'])})
    return report



from app.lookups.sqlite.base import DatabaseConnection


class SearchSpaceDB(DatabaseConnection):
    def add_tables(self):
        """Creates a searchspace lookup sqlite."""
        self.create_tables(['known_searchspace', 'protein_peptides'])

    def write_peps(self, peps, reverse_seqs):
        """Writes peps to db. We can reverse to be able to look up
        peptides that have some amino acids missing at the N-terminal.
        This way we can still use the index.
        """
        if reverse_seqs:
            peps = [(x[0][::-1],) for x in peps]
        cursor = self.get_cursor()
        cursor.executemany(
            'INSERT INTO known_searchspace(seqs) VALUES (?)', peps)
        self.conn.commit()

    def index_peps(self, reverse_seqs):
        if reverse_seqs:
            self.index_column('reverse_seqs_index', 'known_searchspace',
                              'seqs COLLATE NOCASE')
        else:
            self.index_column('reverse_seqs_index', 'known_searchspace',
                              'seqs')

    def store_pep_proteins(self, pepproteins):
        cursor = self.get_cursor()
        cursor.executemany('INSERT INTO protein_peptides(seq, protid, pos) '
                           'VALUES(?, ?, ?)', pepproteins)
        self.conn.commit()

    def index_proteins(self):
        self.index_column('pepix', 'protein_peptides', 'seq')
        self.conn.commit()

    def check_seq_exists(self, seq, amount_ntermwildcards):
        """Look up sequence in sqlite DB. Returns True or False if it
        exists (or not). When looking up a reversed DB with
        ntermwildcards: we reverse the sequence of the pep and add
        a LIKE and %-suffix to the query.
        """
        cursor = self.get_cursor()
        if amount_ntermwildcards > 0:
            seq = seq[::-1]
            sqlseq = '{}{}'.format(seq, '%')
            # FIXME non-parametrized string binding because if ? binding is
            # used the INDEX is not used when looking up, apparently because
            # the query cant be optimized when using LIKE and binding.
            sql = ('select seqs from known_searchspace where seqs LIKE '
                   '"{}"'.format(sqlseq))
            for match in cursor.execute(sql):
                if match[0][:-amount_ntermwildcards] in seq:
                    return True
            return False
        else:
            sql = ('select exists(select seqs from known_searchspace '
                   'where seqs=? limit 1)')
            return cursor.execute(sql, (seq, )).fetchone()[0] == 1

    def get_protein_from_pep(self, peptide):
        cursor = self.get_cursor()
        cursor.execute('SELECT protid, pos FROM protein_peptides WHERE seq='
                       '"{}"'.format(peptide))
        return cursor



from statistics import median, StatisticsError


def get_medians(channels, ratios):
    ch_medians = {}
    for ix, channel in enumerate(channels):
        try:
            ch_medians[channel] = median([x[ix] for x in ratios
                                          if x[ix] != 'NA'])
        except StatisticsError:
            # channel is empty, common in protein quant but not in normalizing
            ch_medians[channel] = 'NA'
    return ch_medians



from app.lookups.sqlite.base import ResultLookupInterface


class BioSetDB(ResultLookupInterface):
    def add_tables(self):
        self.create_tables(['biosets', 'mzmlfiles'])

    def store_biosets(self, biosets):
        self.store_many('INSERT INTO biosets(set_name) VALUES (?)', biosets)

    def store_mzmlfiles(self, mzmlfiles):
        self.store_many('INSERT INTO mzmlfiles(mzmlfilename, set_id) '
                        'VALUES (?, ?)', mzmlfiles)

    def index_biosets(self):
        self.index_column('setname_index', 'biosets', 'set_name')
        self.index_column('setid_index', 'biosets', 'set_id')
        self.index_column('setid_mzmlfn_index', 'mzmlfiles', 'set_id')
        self.index_column('mzmlfn_index', 'mzmlfiles', 'mzmlfilename')
        self.index_column('mzmlfnid_index', 'mzmlfiles', 'mzmlfile_id')

    def get_setnames(self):
        cursor = self.get_cursor()
        cursor.execute('SELECT set_name, set_id FROM biosets')
        return {setname: set_id for setname, set_id in cursor.fetchall()}



from app.drivers.pycolator import base
from app.actions.pycolator import filters as preparation
from app.drivers.options import pycolator_options


class FilterPeptideLength(base.PycolatorDriver):
    """Filters on peptide length, to be specified in calling. Outputs to
    multiple files if multiple file input is given. No PSMs will be
    outputted."""
    outsuffix = '_filtlen.xml'
    command = 'filterlen'
    commandhelp = 'Filters out peptides that exceed --maxlen and --minlen'

    def set_options(self):
        super().set_options()
        options = self.define_options(['maxlength', 'minlength'],
                                      pycolator_options)
        self.options.update(options)

    def set_features(self):
        # FIXME psm filter len too!
        self.features = {
            'psm': preparation.filter_peptide_length(
                self.allpsms, 'psm', self.ns, self.minlength, self.maxlength),
            'peptide': preparation.filter_peptide_length(
                self.allpeps, 'pep', self.ns, self.minlength, self.maxlength)
        }


class FilterUniquePeptides(base.PycolatorDriver):
    """This class processes multiple percolator runs from fractions and
    filters out the best scoring peptides."""
    outsuffix = '_filtuni.xml'
    command = 'filteruni'
    commandhelp = ('Only includes best scoring unique peptides in a (merged) '
                   'file')

    def set_options(self):
        super().set_options()
        self.options.update(self.define_options(['score'], pycolator_options))

    def get_all_psms(self):
        """Override parent method so it returns strings instead"""
        return self.get_all_psms_strings()

    def set_features(self):
        uniquepeps = preparation.filter_unique_peptides(self.allpeps,
                                                        self.score,
                                                        self.ns)
        self.features = {'psm': self.allpsms, 'peptide': uniquepeps}


class FilterWholeProteinSequence(base.PycolatorDriver):
    """This class processes multiple percolator runs from fractions and
    filters out first peptides that are found in a specified searchspace. Then
    it keeps the remaining best scoring unique peptides."""
    outsuffix = '_filtprot.xml'
    command = 'filterprot'
    lookuptype = 'searchspace'
    commandhelp = ('Filters out peptides that match to a protein in a FASTA '
                   'file. Needs a lookup to match peptides to quickly.')

    def set_options(self):
        super().set_options()
        self.options.update(self.define_options(['deamidate', 'fasta',
                                                 'minlength', 'lookupfn',
                                                 'forcetryp'],
                                                pycolator_options))

    def set_features(self):
        self.features = {
            'peptide': preparation.filter_whole_proteins(self.allpeps,
                                                         self.fasta,
                                                         self.lookup,
                                                         'pep',
                                                         self.ns,
                                                         self.deamidate,
                                                         self.minlength,
                                                         self.forcetryp),
            'psm': preparation.filter_whole_proteins(self.allpsms,
                                                     self.fasta,
                                                     self.lookup,
                                                     'psm',
                                                     self.ns,
                                                     self.deamidate,
                                                     self.minlength,
                                                     self.forcetryp)
        }


class FilterPeptideSequence(base.PycolatorDriver):
    """This class processes multiple percolator runs from fractions and
    filters out first peptides that are found in a specified searchspace. Then
    it keeps the remaining best scoring unique peptides."""
    outsuffix = '_filtseq.xml'
    lookuptype = 'searchspace'
    command = 'filterseq'
    commandhelp = ('Filters out peptides that are found in a certain lookup '
                   'DB')

    def set_options(self):
        super().set_options()
        self.options.update(self.define_options(['falloff', 'deamidate',
                                                 'lookupfn'],
                                                pycolator_options))

    def set_features(self):
        self.features = {
            'peptide': preparation.filter_known_searchspace(self.allpeps,
                                                            'pep',
                                                            self.lookup,
                                                            self.ns,
                                                            self.falloff,
                                                            self.deamidate),
            'psm': preparation.filter_known_searchspace(self.allpsms,
                                                        'psm',
                                                        self.lookup,
                                                        self.ns,
                                                        self.falloff,
                                                        self.deamidate),
        }



import os
from argparse import ArgumentParser, RawTextHelpFormatter

VERSION_NUMBER = '2.12'


def parser_file_exists(currentparser, fn):
    if not os.path.exists(fn):
        currentparser.error('Input file %s not found!' % fn)
    else:
        return fn


def populate_parser(drivers):
    parser = ArgumentParser(formatter_class=RawTextHelpFormatter)
    parser.add_argument('--version', action='version', version=VERSION_NUMBER)
    subparsers = parser.add_subparsers(dest='subcommand')
    subparsers.required = True
    subparsermap = {}
    for driver in drivers:
        driver.set_options()
        cmd = driver.get_commandname()
        subparsermap[cmd] = subparsers.add_parser(
            cmd, conflict_handler='resolve', help=driver.get_commandhelp())
        subparsermap[cmd].set_defaults(func=driver.start)
        subparsermap[cmd].add_argument('--version', action='version',
                                       version=VERSION_NUMBER)
        options = [{k: v for k, v in opt.items()}
                   for opt in driver.get_options()]
        for argoptions in options:
            argoptions['dest'] = argoptions.pop('driverattr')
            clarg = argoptions.pop('clarg')
            if 'type' in argoptions and argoptions['type'] == 'file':
                argoptions['type'] = lambda x: parser_file_exists(parser, x)
            elif 'type' in argoptions and argoptions['type'] == 'pick':
                argoptions.pop('picks')
                argoptions.pop('type')
            if not 'required' in argoptions:
                argoptions['required'] = True
            if type(clarg) == list:
                subparsermap[cmd].add_argument(*clarg, **argoptions)
            else:
                subparsermap[cmd].add_argument(clarg, **argoptions)
    return parser


def start_msstitch(exec_drivers, sysargs):
    """Passed all drivers of executable, checks which command is passed to
    the executable and then gets the options for a driver, parses them from
    command line and runs the driver"""
    parser = populate_parser(exec_drivers)
    args = parser.parse_args(sysargs[1:])
    args.func(**vars(args))



from app.drivers.mslookup import base
from app.actions.mslookup import searchspace as preparation
from app.drivers.options import mslookup_options


class SeqspaceLookupDriver(base.LookupDriver):
    """Creates an SQLite lookup DB from a FASTA file. Sequences are
    trypsinized and stored. It's possible to store sequences reversed
    for N-terminal falloff indexing, and it can be specified to cut
    tryptic before proline.
    """
    lookuptype = 'searchspace'
    command = 'seqspace'
    commandhelp = """Create a lookup DB from a FASTA file. Sequences are
    trypsinized and stored. It's possible to store sequences reversed
    for N-terminal falloff indexing, and it can be specified to cut
    tryptic before proline."""

    def __init__(self):
        super().__init__()
        self.infiletype = 'FASTA'

    def set_options(self):
        super().set_options()
        self.options['--dbfile'].update({'required': False, 'default': None})
        self.options.update(self.define_options(['falloff', 'proline',
                                                 'trypsinize'],
                                                mslookup_options))

    def create_lookup(self):
        preparation.create_searchspace(self.lookup, self.fn, self.proline,
                                       self.falloff, self.trypsinize)


class WholeProteinSeqspaceLookupDriver(base.LookupDriver):
    lookuptype = 'searchspace'
    command = 'protspace'
    commandhelp = """Create a full-length protein lookup DB from a FASTA file.
    Protein sequences are stored as short peptides starting from every one of
    the proteins' amino acids."""

    def __init__(self):
        super().__init__()
        self.infiletype = 'FASTA'

    def set_options(self):
        super().set_options()
        self.options['--dbfile'].update({'required': False, 'default': None})
        self.options.update(self.define_options(['minlength'],
                                                mslookup_options))

    def create_lookup(self):
        preparation.create_searchspace_wholeproteins(self.lookup, self.fn,
                                                     self.minlength)



from app.readers import tsv as reader
from app.readers import fasta
from app.dataformats import prottable as prottabledata
from app.dataformats import mzidtsv as mzidtsvdata


def generate_protein_fdr(target, decoy, theader, dheader, headerfields):
    tproteins = [x for x in reader.generate_tsv_proteins(target, theader)]
    dproteins = [x for x in reader.generate_tsv_proteins(decoy, dheader)]
    [x.update({'target_decoy': 'target'}) for x in tproteins]
    [x.update({'target_decoy': 'decoy'}) for x in dproteins]
    proteins = sorted(tproteins + dproteins, key=lambda x: get_score(x),
                      reverse=True)
    fdrheader = headerfields['proteinfdr'][prottabledata.HEADER_QVAL][None]
    for protein in qvalue_generator(fdrheader, proteins):
        yield protein


def generate_pick_fdr(target, decoy, theader, dheader, headerfields,
                      targetfasta, decoyfasta, inferencetype,
                      fastadelim, genefield):
    picked_proteins = get_picked_proteins(target, decoy, theader, dheader,
                                          targetfasta, decoyfasta,
                                          inferencetype, fastadelim, genefield)
    sorted_proteins = sorted(picked_proteins, key=lambda x: get_score(x),
                             reverse=True)
    fdrheader = headerfields['proteinfdr'][prottabledata.HEADER_QVAL][None]
    for protein in qvalue_generator(fdrheader, sorted_proteins):
        yield protein


def qvalue_generator(fdrheader, sorted_features):
    tdcounter = {'target': 0, 'decoy': 0}
    for feat in sorted_features:
        outfeat = {k: v for k, v in feat.items()}
        tdcounter[feat['target_decoy']] += 1
        fdr = tdcounter['decoy'] / float(tdcounter['decoy'] +
                                         tdcounter['target'])
        if outfeat['target_decoy'] == 'target':
            outfeat[fdrheader] = fdr
            yield outfeat


def get_picked_proteins(targetprot, decoyprot, theader, dheader, targetfasta,
                        decoyfasta, picktype, fastadelim, genefield):
    t_feats, d_feats = generate_accessionmaps(targetprot, decoyprot, theader,
                                              dheader)
    if picktype == 'fasta':
        tdmap = create_td_gene_map(targetfasta, decoyfasta,
                                   fastadelim, genefield)
    elif picktype == 'result':
        tdmap = create_td_assoc_map(targetprot, decoyprot, theader, dheader)
    picked_proteins = []
    for tgene, dgene in tdmap.items():
        picked = pick_target_decoy(t_feats.get(tgene), d_feats.get(dgene))
        if picked:
            picked_proteins.append(picked)
    return picked_proteins


def get_score(protein):
    try:
        return float(protein[prottabledata.HEADER_QSCORE])
    except (TypeError, ValueError):
        return False


def create_td_gene_map(tfastafn, dfastafn, fastadelim, genefield):
    tfasta = (x[1].split('.')[0] for x in
              fasta.get_proteins_genes(tfastafn, fastadelim, genefield))
    dfasta = (x[1].split('.')[0] for x in
              fasta.get_proteins_genes(dfastafn, fastadelim, genefield))
    tdmap = {}
    for target, decoy in zip(tfasta, dfasta):
        tdmap[target] = decoy
    return tdmap


def create_td_assoc_map(tproteins, dproteins, theader, dheader):
    tdmap = {}
    for dprot in reader.generate_tsv_proteins(dproteins, dheader):
        dprot = dprot[prottabledata.HEADER_PROTEIN]
        faketprot = dprot.replace(mzidtsvdata.DECOY_PREFIX, '')
        tdmap[faketprot] = dprot
    for tprot in reader.generate_tsv_proteins(tproteins, theader):
        tprot = tprot[prottabledata.HEADER_PROTEIN]
        if not tdmap.get(tprot, False):
            tdmap[tprot] = None
    return tdmap


def generate_accessionmaps(targetprot, decoyprot, theader, dheader):
    t_scores, d_scores = {}, {}
    for protein in reader.generate_tsv_proteins(targetprot, theader):
        acc = protein[prottabledata.HEADER_PROTEIN]
        t_scores[acc] = protein
        t_scores[acc]['target_decoy'] = 'target'
    for protein in reader.generate_tsv_proteins(decoyprot, dheader):
        acc = protein[prottabledata.HEADER_PROTEIN]
        d_scores[acc] = protein
        d_scores[acc]['target_decoy'] = 'decoy'
    return t_scores, d_scores


def pick_target_decoy(tfeature, dfeature):
    """Feed it with a target and decoy score and the protein/gene/id names,
    and this will return target/decoy type, the winning ID and the score"""
    tscore, dscore = get_score(tfeature), get_score(dfeature)
    if tscore == dscore:
        # same score or both False
        return False
    elif False in [tscore, dscore]:
        # return the non-False feature
        return [v for k, v in {tscore: tfeature, dscore: dfeature}.items()
                if k is not False][0]
    elif tscore > dscore:
        return tfeature
    elif tscore < dscore:
        return dfeature
    else:
        # in case uncaught edgecase occurs
        print('WARNING, target score {} and decoy score {} could not be '
              'compared'.format(tscore, dscore))
        return False



import os
from numpy import polyfit
from math import log

from tests.integration import basetests


class TestPSM2Peptable(basetests.PeptableTest):
    command = 'psm2pep'
    suffix = '_peptable.tsv'
    infilename = 'mzidtsv.txt'

    def check(self, fncol):
        psms = {}
        for psm in self.tsv_generator(self.infile[0]):
            score = str(float(psm['percolator svm-score']))
            try:
                psms[psm['Peptide']][score] = psm
            except KeyError:
                psms[psm['Peptide']] = {score: psm}
        striplist = ['fake_ch{}'.format(x) for x in range(8)]
        striplist.extend(['Files/scans for peptide', 'MS1 area',
                          'MS1 area (highest of all PSMs)'])
        for peptide in self.tsv_generator(self.resultfn):
            for newkey, oldkey in zip(['Peptide', 'peptide PEP',
                                       'peptide q-value', 'Protein'],
                                      ['Peptide sequence', 'PEP', 'q-value',
                                       'Protein(s)']):
                peptide[newkey] = peptide.pop(oldkey)
            mapping_psms = psms[peptide['Peptide']]
            toppsm = mapping_psms[str(max([float(score) for score
                                           in mapping_psms.keys()]))]
            testpeptide = {k: v for k, v in peptide.items()
                           if k not in striplist}
            testpsm = {k: v for k, v in toppsm.items()
                       if k not in striplist}
            self.assertEqual(testpeptide, testpsm)
            ms1s = [self.get_float_or_na(psm['MS1 area'])
                    for psm in mapping_psms.values()]
            try:
                topms1 = max(ms1s)
            except TypeError:
                # cannot do max on mix of str and float
                topms1 = max([0 if x == 'NA' else x for x in ms1s])
            self.assertEqual(str(topms1),
                             peptide['MS1 area (highest of all PSMs)'])
            psmfnscans = set(peptide['Files/scans for peptide'].split('; '))
            with open(self.infile[0]) as fp:
                fnfield = next(fp).strip().split('\t')[fncol - 1]
            self.assertEqual(psmfnscans,
                             set(['{}_{}'.format(psm[fnfield], psm['ScanNum'])
                                  for psm in mapping_psms.values()]))

    def test_psm2peptable(self):
        fncol = 1
        options = ['--spectracol', str(fncol), '--isobquantcolpattern',
                   'fake_ch', '--scorecolpattern', 'svm',
                   '--ms1quantcolpattern', 'MS1']
        self.run_command(options)
        self.check(fncol)

    def test_no_spectracol(self):
        options = ['--isobquantcolpattern', 'fake_ch',
                   '--scorecolpattern', 'svm', '--ms1quantcolpattern', 'MS1']
        self.run_command(options)
        self.check(1)


class TestIsoquant(basetests.PeptableTest):
    infilename = 'peptable_small.txt'
    command = 'isoquant'
    suffix = '_isoq.txt'
    channels = ['fake_ch{}'.format(x) for x in range(0, 8)]
    nopsms = ['{} - # quanted PSMs'.format(ch) for ch in channels]

    def test_isoquant(self):
        isotable = os.path.join(self.fixdir, 'peptable_isoquant.txt')
        options = ['--quantfile', isotable, '--isobquantcolpattern', 'fake_ch',
                   '--qaccpattern', 'Protein ID']
        self.run_command(options)
        self.isoquant_check(isotable, 'Peptide sequence', self.channels,
                            self.nopsms)


class TestModelQvals(basetests.PeptableTest):
    command = 'modelqvals'
    infilename = 'peptable_small.txt'
    suffix = '_qmodel.txt'

    def test_modelqvals(self):
        score, fdr = 'percolator svm-score', '^q-value'
        options = ['--scorecolpattern', score, '--fdrcolpattern', fdr]
        self.run_command(options)
        scores, qvalues = [], []
        for line in self.tsv_generator(self.infile[0]):
            if float(line[fdr[1:]]) > 10e-4:
                scores.append(float(line[score]))
                qvalues.append(log(float(line[fdr[1:]]), 10))
        slope, intercept = polyfit(scores, qvalues, deg=1)
        for line in self.tsv_generator(self.resultfn):
            self.assertAlmostEqual(float(line[fdr[1:] + ' (linear modeled)']),
                                   10 ** (float(line[score]) *
                                          slope + intercept))


class TestBuild(basetests.PeptableTest):
    command = 'build'
    infilename = 'built_peptide_table.txt'
    suffix = ''

    def setUp(self):
        super().setUp()
        self.dbfile = os.path.join(self.fixdir, 'peptable_db.sqlite')

    def get_std_options(self):
        cmd = [self.executable, self.command, '-d', self.workdir]
        return cmd

    def check(self, genecentric=False, noncentric=False):
        self.check_values()
        if genecentric:
            self.check_genes()
        elif noncentric:
            pass
        else:
            self.check_proteingroups()
        self.check_isobaric()

    def check_values(self):
        sql = ('SELECT ps.sequence, bs.set_name, '
               'ppq.quant, pf.fdr, pp.pep '
               'FROM peptide_sequences AS ps '
               'JOIN biosets AS bs '
               'JOIN peptide_precur_quanted AS ppq USING(pep_id) '
               'JOIN peptide_fdr AS pf USING(pep_id) '
               'JOIN peptide_pep AS pp USING(pep_id) '
               )
        self.check_build_values(sql, ['MS1 area (highest of all PSMs)',
                                      'q-value', 'PEP'], 'Peptide sequence')

    def check_genes(self):
        # FIXME get a db WITHOUT protein groups for genecentric peptides
        sql = ('SELECT ps.sequence, p.psm_id, "NA", pd.description, '
               'g.gene_acc, aid.assoc_id, "NA" '
               'FROM peptide_sequences AS ps '
               'JOIN psms AS p USING(pep_id) '
               'JOIN protein_psm USING(psm_id) '
               'JOIN prot_desc AS pd USING(protein_acc) '
               'JOIN genes AS g USING(protein_acc) '
               'JOIN associated_ids AS aid USING(protein_acc) '
               'JOIN protein_coverage AS pc USING(protein_acc) '
               )
        self.check_peptide_relations(sql)

    def check_proteingroups(self):
        # FIXME have to also check content proteins, etc
        sql = ('SELECT ps.sequence, p.psm_id, pm.protein_acc, pd.description, '
               'g.gene_acc, aid.assoc_id, pc.coverage '
               'FROM peptide_sequences AS ps '
               'JOIN psms AS p USING(pep_id) '
               'JOIN psm_protein_groups USING(psm_id) '
               'JOIN protein_group_master AS pm USING(master_id) '
               'JOIN prot_desc AS pd USING(protein_acc) '
               'JOIN genes AS g USING(protein_acc) '
               'JOIN associated_ids AS aid USING(protein_acc) '
               'JOIN protein_coverage AS pc USING(protein_acc) '
               )
        self.check_peptide_relations(sql)

    def check_peptide_relations(self, sql):
        expected, psm_id, pep = {}, None, None
        for rec in self.get_values_from_db(self.dbfile, sql):
            try:
                expected[rec[0]]['psms'].add(rec[1])
            except KeyError:
                expected[rec[0]] = {'psms': set([rec[1]]),
                                    'pgroups': set([rec[2]]),
                                    'descriptions': set([rec[3]]),
                                    'genes': set([rec[4]]),
                                    'assoc': set([rec[5]]),
                                    'cover': set([str(rec[6])]),
                                    }
            else:
                expected[rec[0]]['pgroups'].add(rec[2])
                expected[rec[0]]['descriptions'].add(rec[3])
                expected[rec[0]]['genes'].add(rec[4])
                expected[rec[0]]['assoc'].add(rec[5])
                expected[rec[0]]['cover'].add(str(rec[6]))
        for line in self.tsv_generator(self.resultfn):
            self.assertEqual(set(line['Protein(s)'].split(';')),
                             expected[line['Peptide sequence']]['pgroups'])
            self.assertEqual(
                set(line['Description(s)'].split(';')),
                set([y for x in
                     expected[line['Peptide sequence']]['descriptions']
                     for y in x.split(';')]))
            self.assertEqual(set(line['Gene(s)'].split(';')),
                             expected[line['Peptide sequence']]['genes'])
            self.assertEqual(set(line['Associated gene ID(s)'].split(';')),
                             expected[line['Peptide sequence']]['assoc'])
            self.assertEqual(set([line['Coverage(s)']]),
                             expected[line['Peptide sequence']]['cover'])

    def check_isobaric(self):
        sql = ('SELECT ps.sequence, bs.set_name, ch.channel_name, '
               'iq.quantvalue, iq.amount_psms '
               'FROM peptide_sequences AS ps '
               'JOIN biosets AS bs '
               'JOIN peptide_iso_quanted AS iq USING(pep_id) '
               'JOIN pepquant_channels AS ch USING(channel_id) '
               )
        self.check_built_isobaric(sql, 'Peptide sequence')

    def test_proteincentric(self):
        options = ['--fdr', '--pep', '--isobaric', '--precursor',
                   '--dbfile', self.dbfile]
        self.run_command(options)
        self.check()

    def test_genecentric(self):
        options = ['--fdr', '--pep', '--isobaric', '--precursor',
                   '--dbfile', self.dbfile, '--genecentric']
        self.run_command(options)
        self.check(genecentric=True)

    def test_noncentric(self):
        options = ['--fdr', '--pep', '--isobaric', '--precursor',
                   '--dbfile', self.dbfile, '--noncentric']
        self.run_command(options)
        self.check(noncentric=True)



from app.drivers.mslookup import base
from app.actions.mslookup import biosets as lookups
from app.drivers.options import mslookup_options


class BioSetLookupDriver(base.LookupDriver):
    outsuffix = '_setlookup.sqlite'
    lookuptype = 'biosets'
    command = 'biosets'
    commandhelp = ('Create SQLite lookup of mzML input files and '
                   'biological set names. Input files are passed to -i, '
                   'respective set names are passed to --setnames.')

    def __init__(self):
        super().__init__()
        self.infiletype = 'mzML spectra'

    def set_options(self):
        super().set_options()
        del(self.options['--dbfile'])
        self.options.update(self.define_options(['setnames'],
                                                mslookup_options))

    def parse_input(self, **kwargs):
        super().parse_input(**kwargs)
        self.lookupfn = None
        self.setnames = [x.replace('"', '') for x in self.setnames]

    def create_lookup(self):
        lookups.create_bioset_lookup(self.lookup, self.fn, self.setnames)



from app.writers import tsv


def write_prottable(header, psms, outfn):
    tsv.write_tsv(header, psms, outfn)


def write_prottable_with_na(header, psms, outfn):
    na_psms = (tsv.parse_NA(x, header) for x in psms)
    write_prottable(header, na_psms, outfn)



import os


def create_bioset_lookup(lookupdb, spectrafns, set_names):
    """Fills lookup database with biological set names"""
    unique_setnames = set(set_names)
    lookupdb.store_biosets(((x,) for x in unique_setnames))
    set_id_map = lookupdb.get_setnames()
    mzmlfiles = ((os.path.basename(fn), set_id_map[setname])
                 for fn, setname in zip(spectrafns, set_names))
    lookupdb.store_mzmlfiles(mzmlfiles)
    lookupdb.index_biosets()



from app.readers.mzidplus import (PERCO_ORDER, P_PEPTIDEQ, P_PEPTIDEPEP)

DECOY_PREFIX = 'decoy_'

HEADER_SPECFILE = '#SpecFile'
HEADER_SCANNR = 'ScanNum'
HEADER_CHARGE = 'Charge'
HEADER_PEPTIDE = 'Peptide'
HEADER_PROTEIN = 'Protein'
HEADER_GENE = 'Gene'
HEADER_SYMBOL = 'Gene Symbol'
HEADER_DESCRIPTION = 'Description'
HEADER_PRECURSOR_MZ = 'Precursor'
HEADER_PRECURSOR_QUANT = 'MS1 area'
HEADER_MASTER_PROT = 'Master protein(s)'
HEADER_PG_CONTENT = 'Protein group(s) content'
HEADER_PG_AMOUNT_PROTEIN_HITS = 'Amount of matching proteins in group(s)'
HEADER_PG = [HEADER_MASTER_PROT, HEADER_PG_CONTENT,
             HEADER_PG_AMOUNT_PROTEIN_HITS]
HEADER_PEPTIDE_Q = P_PEPTIDEQ[1]
HEADER_PEPTIDE_PEP = P_PEPTIDEPEP[1]
HEADER_MSGFSCORE = 'MSGFScore'
HEADER_EVALUE = 'EValue'
HEADER_SETNAME = 'Biological set'
HEADER_RETENTION_TIME = 'Retention time(min)'
HEADER_INJECTION_TIME = 'Ion injection time(ms)'
SPECDATA_HEADER = [HEADER_SETNAME, HEADER_RETENTION_TIME, HEADER_INJECTION_TIME]
PERCO_HEADER = [x[1] for x in PERCO_ORDER]



from collections import OrderedDict

from app.readers import tsv
from app.actions.headers.base import (generate_general_header,
                                      generate_headerfields,
                                      get_isoquant_fields)
from app.dataformats import peptable as peptabledata
from app.dataformats import mzidtsv as mzidtsvdata


def switch_psm_to_peptable_fields(oldheader):
    """Returns a dict map with old to new header fields"""
    return {old: new for old, new in zip([mzidtsvdata.HEADER_PEPTIDE,
                                          mzidtsvdata.HEADER_PROTEIN,
                                          mzidtsvdata.HEADER_PEPTIDE_Q,
                                          mzidtsvdata.HEADER_PEPTIDE_PEP],
                                         [peptabledata.HEADER_PEPTIDE,
                                          peptabledata.HEADER_PROTEINS,
                                          peptabledata.HEADER_QVAL,
                                          peptabledata.HEADER_PEP])}


def get_psm2pep_header(oldheader, isobq_pattern=False, precurqfield=False):
    header = oldheader[:]
    if isobq_pattern:
        isocols = tsv.get_columns_by_pattern(header, isobq_pattern)
        for col in isocols:
            header.pop(header.index(col))
    if precurqfield:
        header = [peptabledata.HEADER_AREA if x == precurqfield
                  else x for x in header]
    peptable_header = [peptabledata.HEADER_LINKED_PSMS]
    ix = header.index(mzidtsvdata.HEADER_PEPTIDE)
    header = header[:ix] + peptable_header + header[ix:]
    switch_map = switch_psm_to_peptable_fields(header)
    return [switch_map[field] if field in switch_map else field
            for field in header]


def get_linear_model_header(oldheader):
    header = oldheader[:]
    try:
        ix = header.index(peptabledata.HEADER_PEP) + 1
    except ValueError:
        ix = header.index(peptabledata.HEADER_QVAL) + 1
    return header[:ix] + [peptabledata.HEADER_QVAL_MODELED] + header[ix:]


def generate_header(headerfields, oldheader, group_by_field):
    """Returns a header as a list, ready to write to TSV file"""
    fieldtypes = ['peptidefdr', 'peptidepep', 'nopsms', 'proteindata',
                  'precursorquant', 'isoquant']
    return generate_general_header(headerfields, fieldtypes,
                                   peptabledata.HEADER_PEPTIDE, oldheader,
                                   group_by_field)


def get_peptable_headerfields(headertypes, lookup=False, poolnames=False):
    """Called by driver to generate headerfields object"""
    field_defs = {'isoquant': get_isoquant_fields,
                  'precursorquant': get_precursorquant_fields,
                  'peptidefdr': get_peptidefdr_fields,
                  'peptidepep': get_peptidepep_fields,
                  'proteindata': get_proteininfo_fields,
                  }
    return generate_headerfields(headertypes, field_defs, poolnames, lookup)


def get_precursorquant_fields(poolnames=False):
    return {peptabledata.HEADER_AREA: poolnames}


def get_peptidefdr_fields(poolnames=False):
    return {peptabledata.HEADER_QVAL: poolnames}


def get_peptidepep_fields(poolnames=False):
    return {peptabledata.HEADER_PEP: poolnames}


def get_proteininfo_fields(poolnames=False, genecentric=False):
    """Returns header fields for protein (group) information."""
    allfields = OrderedDict()
    basefields = [peptabledata.HEADER_PROTEINS,
                  peptabledata.HEADER_GENES,
                  peptabledata.HEADER_ASSOCIATED,
                  peptabledata.HEADER_DESCRIPTIONS,
                  peptabledata.HEADER_COVERAGES,
                  peptabledata.HEADER_NO_CONTENTPROTEINS,
                  ]
    for field in basefields:
        allfields[field] = False
    allfields[peptabledata.HEADER_NO_PSM] = poolnames
    return allfields



import unittest
import subprocess
import os
import shutil
import sqlite3
import re
from lxml import etree
from tempfile import mkdtemp


class BaseTest(unittest.TestCase):
    testdir = 'tests'
    fixdir = os.path.join(os.getcwd(), testdir, 'fixtures')
    outdir = os.path.join(os.getcwd(), testdir, 'test_output')

    def setUp(self):
        self.infile = os.path.join(self.fixdir, self.infilename)
        os.makedirs(self.outdir, exist_ok=True)
        self.workdir = mkdtemp(dir=self.outdir)
        os.chdir(self.workdir)
        self.resultfn = os.path.join(self.workdir,
                                     self.infilename + self.suffix)

    def tearDown(self):
        shutil.rmtree(self.workdir)

    def get_std_options(self):
        cmd = [self.executable, self.command, '-i']
        if type(self.infile) != list:
            self.infile = [self.infile]
        cmd.extend(self.infile)
        if self.resultfn is not None and self.executable != 'msslookup':
            cmd.extend(['-o', self.resultfn])
        elif self.resultfn is None:
            cmd.extend(['-d', self.workdir])
        return cmd

    def run_command(self, options=[]):
        cmd = self.get_std_options()
        cmd.extend(options)
        try:
            subprocess.check_call(cmd)
        except subprocess.CalledProcessError:
            print('Failed to run executable {}'.format(self.executable))
            raise
        return cmd

    def run_command_expect_error(self, options=[]):
        try:
            cmd = self.run_command(options)
        except subprocess.CalledProcessError:
            pass
        else:
            self.fail('Command {} should throw an error'.format(cmd))

    def run_command_stdout(self, options=[]):
        cmd = self.get_std_options()
        cmd.extend(options)
        try:
            return subprocess.check_output(cmd)
        except subprocess.CalledProcessError:
            print('Failed to run executable {}'.format(self.executable))
            raise

    def get_values_from_db(self, dbfile, sql):
        db = sqlite3.connect(dbfile)
        return db.execute(sql)

    def seq_in_db(self, dbconn, seq, seqtype, max_falloff=False):
        seq = seq.replace('L', 'I')
        if seqtype == 'ntermfalloff':
            seq = '{0}%'.format(seq[::-1])
            sql = 'SELECT seqs FROM known_searchspace WHERE seqs LIKE ?'
            for match in dbconn.execute(sql, (seq,)):
                if match[0][:-max_falloff] in seq:
                    return True
        else:
            sql = ('SELECT EXISTS(SELECT seqs FROM known_searchspace WHERE '
                   'seqs=? LIMIT 1)')
            return dbconn.execute(sql, (seq,)).fetchone()[0] == 1

    def get_tsvheader(self, fn):
        with open(fn) as fp:
            return next(fp).strip('\n').split('\t')

    def tsv_generator(self, fn):
        header = self.get_tsvheader(fn)
        tsv = self.get_all_lines(fn)
        for line in tsv:
            yield {field: val for field, val in
                   zip(header, line.strip('\n').split('\t'))}

    def get_all_lines(self, fn):
        with open(fn) as fp:
            next(fp)
            for line in fp:
                yield line

    def get_xml_namespace(self, fn):
        root = self.get_root_el(fn)
        ns = {}
        for prefix in root.nsmap:
            separator = ':'
            nsprefix = prefix
            if prefix is None:
                nsprefix = ''
                separator = ''
            ns['xmlns{0}{1}'.format(separator, nsprefix)] = root.nsmap[prefix]
        return ns

    def get_root_el(self, fn):
        rootgen = etree.iterparse(fn, events=('start',))
        root = next(rootgen)[1]
        for child in root.getchildren():
            root.remove(child)
        return root

    def copy_db_to_workdir(self, dbfn):
        shutil.copy(os.path.join(self.fixdir, dbfn), self.resultfn)

    def get_float_or_na(self, value):
        try:
            return float(value)
        except ValueError:
            return value

    def check_lines(self, expected, result):
        with open(expected) as fp, open(result) as resultfp:
            for expline, resline in zip(fp, resultfp):
                self.assertEqual(expline, resline)

    def isoquant_check(self, expected_isotable, acc_field, channels, nopsms):
        isoquant = {}
        accession = self.get_tsvheader(expected_isotable)[0]
        for line in self.tsv_generator(expected_isotable):
            acc = line.pop(accession)
            isoquant[acc] = line
        for result in self.tsv_generator(self.resultfn):
            for ch in channels + nopsms:
                try:
                    resval = float(result[ch])
                    expval = float(isoquant[result[acc_field]][ch])
                except ValueError:
                    # NA found
                    self.assertEqual(result[ch], isoquant[result[acc_field]][ch])
                else:
                    self.assertEqual(resval, expval)


class BaseTestPycolator(BaseTest):
    executable = 'msspercolator'
    infilename = 'percolator_out.xml'

    def get_psm_pep_ids_from_file(self, fn):
        contents = self.read_percolator_out(fn)
        return {'psm_ids': self.get_element_ids(contents['psms'],
                                                'psm_id', contents['ns']),
                'peptide_ids': self.get_element_ids(contents['peptides'],
                                                    'peptide_id',
                                                    contents['ns']),
                'psm_seqs': self.get_psm_seqs(contents['psms'],
                                              contents['ns'])
                }

    def read_percolator_out(self, fn):
        ns = self.get_xml_namespace(fn)['xmlns']
        contents = {'ns': ns, 'psms': [], 'peptides': []}
        xml = etree.iterparse(fn)
        for ac, el in xml:
            if el.tag == '{%s}psm' % ns:
                contents['psms'].append(el)
            elif el.tag == '{%s}peptide' % ns:
                contents['peptides'].append(el)
        return contents

    def get_element_ids(self, elements, id_attrib, ns):
        return [x.attrib['{%s}%s' % (ns, id_attrib)] for x in elements]

    def get_psm_seqs(self, psms, ns):
        return [pepseq.attrib['seq'] for pepseq in
                self.get_subelements(psms, 'peptide_seq', ns)]

    def get_svms(self, elements, ns):
        return [svm.text for svm in
                self.get_subelements(elements, 'svm_score', ns)]

    def get_qvals(self, elements, ns):
        return [qval.text for qval in
                self.get_subelements(elements, 'q_value', ns)]

    def get_peps(self, elements, ns):
        return [pep.text for pep in
                self.get_subelements(elements, 'pep', ns)]

    def get_subelements(self, elements, subel, ns):
        return [element.find('{%s}%s' % (ns, subel)) for element in elements]

    def strip_modifications(self, pep):
        return re.sub('\[UNIMOD:\d*\]', '', pep)


class MzidTSVBaseTest(BaseTest):
    executable = 'msspsmtable'
    infilename = 'mzidtsv.txt'

    def setUp(self):
        super().setUp()
        self.dbfile = os.path.join(self.fixdir, 'mzidtsv_db.sqlite')

    def rowify(self, records):
        row, rownr = [], 0
        for record in records:
            if record[0] == rownr:
                row.append(record)
            else:
                yield row
                row = [record]
                rownr += 1

    def check_results(self, checkfields, expected_values):
        for resultvals, exp_vals in zip(self.get_values(checkfields),
                                        expected_values):
            for resultval, expectval in zip(resultvals, exp_vals):
                self.assertEqual([str(x) if x is not None else 'NA'
                                  for x in expectval],
                                 [str(x) for x in resultval])

    def process_dbvalues_both(self, dbfile, sql, channel_fields, permfieldnrs,
                              permfieldnames):
        dbvals = self.get_values_from_db(dbfile, sql)
        outresults, rownr, permvals = [], 0, []
        for record in dbvals:
            if record[0] != rownr:
                for outresult in outresults:
                    yield outresult
                for pfname, pval in zip(permfieldnames, permvals):
                    yield (rownr, pfname, pval)
                if channel_fields != []:
                    outresults = [tuple([record[0]] +
                                        [record[x] for x in channel_fields])]
                permvals = [record[nr] for nr in permfieldnrs]
                rownr += 1
            else:
                permvals = [record[nr] for nr in permfieldnrs]
                if channel_fields != []:
                    result = [record[0]] + [record[x] for x in channel_fields]
                    outresults.append(tuple(result))

    def get_values(self, checkfields):
        with open(self.resultfn) as fp:
            header = next(fp).strip('\n').split('\t')
            fieldindices = [header.index(field) for field in checkfields]
            row = 0
            for line in fp:
                line = line.strip('\n').split('\t')
                if len(checkfields) > 1:
                    yield [(row, field, line[ix]) for field, ix in
                           zip(checkfields, fieldindices)]
                else:
                    yield [(row, line[ix]) for ix in fieldindices]
                row += 1


class MSLookupTest(BaseTest):
    executable = 'msslookup'
    base_db_fn = None
    suffix = ''

    def setUp(self):
        super().setUp()
        self.resultfn = os.path.join(self.workdir, 'mslookup_db.sqlite')
        if self.base_db_fn is not None:
            self.copy_db_to_workdir(self.base_db_fn)

    def run_command(self, options=None):
        if options is None:
            options = []
        if self.base_db_fn is not None:
            options.extend(['--dbfile', self.resultfn])
        super().run_command(options)


class PepProtableTest(BaseTest):
    def check_build_values(self, sql, fields, accession, cutoff=False):
        expected = {}
        for rec in self.get_values_from_db(self.dbfile, sql):
            # skip multiple entries of e.g protein group per peptide
            # or gene per protein
            try:
                expected[rec[0]][rec[1]] = rec[2:]
            except KeyError:
                expected[rec[0]] = {rec[1]: rec[2:]}
        for line in self.tsv_generator(self.resultfn):
            acc = line[accession]
            for setname, pepvals in expected[acc].items():
                for val, field in zip(pepvals, fields):
                    self.assertEqual(str(val),
                                     line['{}_{}'.format(setname, field)])
            expected.pop(acc)
        self.check_exp_empty(expected, cutoff)

    def check_exp_empty(self, expected, cutoff):
        if cutoff:
            self.assertFalse(expected == {})
        else:
            self.assertTrue(expected == {})

    def check_built_isobaric(self, sql, accession, check_nr_psms=True, cutoff=False):
        expected = {}
        for rec in self.get_values_from_db(self.dbfile, sql):
            am_psm = rec[4]
            try:
                expected[rec[0]][rec[1]][rec[2]] = [rec[3], am_psm]
            except KeyError:
                try:
                    expected[rec[0]][rec[1]] = {rec[2]: [rec[3], am_psm]}
                except KeyError:
                    expected[rec[0]] = {rec[1]: {rec[2]: [rec[3], am_psm]}}
            if cutoff:
                expected[rec[0]][rec[1]][rec[2]] = [rec[3], am_psm, rec[5]]
        for line in self.tsv_generator(self.resultfn):
            for setname, fields in expected[line[accession]].items():
                for field, exp_val in fields.items():
                    setfield = '{}_{}'.format(setname, field)
                    if cutoff and exp_val[2] > cutoff:
                        exp_val = ['NA', 'NA']
                    self.assertEqual(line[setfield], str(exp_val[0]))
                    if check_nr_psms:
                        nr_psms = line['{} - # quanted PSMs'.format(setfield)]
                        self.assertEqual(nr_psms, str(exp_val[1]))
            expected.pop(line[accession])
        self.check_exp_empty(expected, cutoff)


class PeptableTest(PepProtableTest):
    executable = 'msspeptable'


class ProttableTest(PepProtableTest):
    executable = 'mssprottable'
    infilename = 'prottable.txt'

    def get_top_psms(self, fn, pepkey, valuekey, lowerbetter=False):
        top_vals = {}
        for psm in self.tsv_generator(fn):
            prot = psm['Master protein(s)']
            seq = psm[pepkey]
            try:
                value = float(psm[valuekey])
            except ValueError:
                continue
            if prot in top_vals and seq in top_vals[prot]:
                if lowerbetter and value > top_vals[prot][seq]:
                    continue
                elif not lowerbetter and value < top_vals[prot][seq]:
                    continue
            try:
                top_vals[prot][seq] = value
            except KeyError:
                top_vals[prot] = {seq: value}
        return top_vals

    def check_protein_data(self, centrictype):
        centric = {'proteincentric': 'pc', 'genecentric': 'gc',
                   'assoccentric': 'ac'}[centrictype]
        if centric == 'pc':
            sql = (
            'SELECT p.protein_acc,GROUP_CONCAT(g.gene_acc),GROUP_CONCAT(aid.assoc_id),pd.description,pcov.coverage '
            'FROM protein_group_master AS p '
            'JOIN associated_ids AS aid USING(protein_acc) '
            'JOIN genes AS g USING(protein_acc) '
            'JOIN prot_desc AS pd USING(protein_acc) '
            'JOIN protein_coverage AS pcov USING(protein_acc) '
            'GROUP BY g.gene_acc'
            )
        elif centric == 'gc':
            sql = (
            'SELECT g.gene_acc, GROUP_CONCAT(aid.assoc_id),GROUP_CONCAT(p.protein_acc),pd.description '
            'FROM genes AS g '
            'JOIN associated_ids AS aid USING(protein_acc) '
            'JOIN proteins AS p USING(protein_acc) '
            'JOIN prot_desc AS pd USING(protein_acc) '
            'GROUP BY g.gene_acc'
            )
        elif centric == 'ac':
            sql = (
            'SELECT aid.assoc_id, GROUP_CONCAT(g.gene_acc),GROUP_CONCAT(p.protein_acc),pd.description '
            'FROM associated_ids AS aid '
            'JOIN genes AS g USING(protein_acc) '
            'JOIN proteins AS p USING(protein_acc) '
            'JOIN prot_desc AS pd USING(protein_acc) '
            'GROUP BY aid.assoc_id'
            )
        expected = {rec[0]: rec[1:] for rec in
                    self.get_values_from_db(self.dbfile, sql)}
        pdatalup = {
                'pc': {'acc': 'Protein ID', 'fields': [('Gene ID', 0), ('Gene Name', 1), ('Coverage', 3)]},
                'gc': {'acc': 'Gene ID', 'fields': [('Gene Name', 0), ('Protein ID(s)', 1)]},
                'ac': {'acc': 'Gene Name', 'fields': [('Gene ID', 0), ('Protein ID(s)', 1)]},
                }
        for row in self.tsv_generator(self.resultfn):
            acc = row[pdatalup[centric]['acc']]
            for (field, ix) in pdatalup[centric]['fields']:
                self.assertEqual(set(str(row[field]).split(';')), set(str(expected[acc][ix]).split(',')))
            self.assertEqual(row['Description'], expected[acc][2])

        psm_sql = ('SELECT {0}.{1}, pp.psm_id, ps.sequence '
                   'FROM {2} AS {0} '
                   'JOIN protein_psm AS pp USING(protein_acc) '
                   'JOIN psms USING(psm_id) '
                   'JOIN peptide_sequences AS ps USING(pep_id) '
                   )
        psm_sql_map = {'pc': ('pgm', 'protein_acc', 'protein_group_master'),
                       'gc': ('g', 'gene_acc', 'genes'),
                       'ac': ('aid', 'assoc_id', 'associated_ids')}[centric]
        psm_sql = psm_sql.format(*psm_sql_map)
        expected, unipeps = {}, {}
        for rec in self.get_values_from_db(self.dbfile, psm_sql):
            pacc = rec[0]
            try:
                unipeps[rec[2]].add(pacc)
            except KeyError:
                unipeps[rec[2]] = set([pacc])
            try:
                expected[pacc]['psms'].add(rec[1])
            except KeyError:
                expected[pacc] = {'psms': set([rec[1]]), 'pep': set([rec[2]]),
                                  'unipep': 0}
            else:
                expected[pacc]['pep'].add(rec[2])
        for pep, prot in unipeps.items():
            if len(prot) == 1:
                expected[prot.pop()]['unipep'] += 1
        for protein in self.tsv_generator(self.resultfn):
            pacc = protein[pdatalup[centric]['acc']]
            poolname = 'S1'
            self.assertEqual(protein['{}_# Unique peptides'.format(poolname)],
                             str(expected[pacc]['unipep']))
            self.assertEqual(protein['{}_# Peptides'.format(poolname)],
                             str(len(expected[pacc]['pep'])))
            self.assertEqual(protein['{}_# PSMs'.format(poolname)],
                             str(len(expected[pacc]['psms'])))



def get_scan_nr(element, attribname):
    """General method to get a scan nr from xml element of mzML/mzIdentML"""
    info = element.attrib[attribname]
    infomap = {y[0]: y[1] for y in [x.split('=') for x in info.split()]}
    return infomap['scan']



from app.lookups.sqlite import proteingroups as lookups


def get_sortfnxs(evidence):
    sortfnxs = [sort_pgroup_peptides,
                sort_pgroup_psms,
                sort_pgroup_score,
                ]
    if evidence:
        sortfnxs.append(sort_evidence_score)
    sortfnxs.extend([sort_pgroup_coverage, sort_alphabet])
    return sortfnxs


def sort_to_get_master(pgroup, evidence):
    sortfnxs = get_sortfnxs(evidence)
    sorted_pg = sort_protein_group(pgroup, sortfnxs, 0)
    return {'master_id': sorted_pg[0][lookups.MASTER_INDEX],
            'protein_acc': sorted_pg[0][lookups.PROTEIN_ACC_INDEX]}


def sort_protein_groups(pgroups, evidence):
    """Gets a protein groups containing dict pgroups, for each master (key)
    there is a list of protein tuples that comprise a proteingroup. This loops
    the groups and returns a sorted group. Assumes master is already sorted
    as the proper master"""
    sortfnxs = get_sortfnxs(evidence)
    pgroups_out = {}
    for master, pgroup in pgroups.items():
        sorted_pg = sort_protein_group(pgroup, sortfnxs, 0)
        pgroups_out[master] = sorted_pg
    return pgroups_out


def sort_protein_group(pgroup, sortfunctions, sortfunc_index):
    """Recursive function that sorts protein group by a number of sorting
    functions."""
    pgroup_out = []
    subgroups = sortfunctions[sortfunc_index](pgroup)
    sortfunc_index += 1
    for subgroup in subgroups:
        if len(subgroup) > 1 and sortfunc_index < len(sortfunctions):
            pgroup_out.extend(sort_protein_group(subgroup,
                                                 sortfunctions,
                                                 sortfunc_index))
        else:
            pgroup_out.extend(subgroup)
    return pgroup_out


def sort_amounts(proteins, sort_index):
    """Generic function for sorting peptides and psms. Assumes a higher
    number is better for what is passed at sort_index position in protein."""
    amounts = {}
    for protein in proteins:
        amount_x_for_protein = protein[sort_index]
        try:
            amounts[amount_x_for_protein].append(protein)
        except KeyError:
            amounts[amount_x_for_protein] = [protein]
    return [v for k, v in sorted(amounts.items(), reverse=True)]


def sort_pgroup_peptides(proteins):
    return sort_amounts(proteins, lookups.PEPTIDE_COUNT_INDEX)


def sort_pgroup_psms(proteins):
    return sort_amounts(proteins, lookups.PSM_COUNT_INDEX)


def sort_pgroup_score(proteins):
    return sort_amounts(proteins, lookups.PROTEIN_SCORE_INDEX)


def sort_pgroup_coverage(proteins):
    return sort_amounts(proteins, lookups.COVERAGE_INDEX)


def sort_evidence_score(proteins):
    return sort_amounts(proteins, lookups.EVIDENCE_LVL_INDEX)


def sort_alphabet(proteins):
    return [sorted(proteins, key=lambda x: x[2])]



from app.actions.mslookup import proteingrouping as lookups
from app.drivers.mslookup import base


class ProteinGroupLookupDriver(base.LookupDriver):
    lookuptype = 'proteingroups'
    command = 'proteingroup'
    commandhelp = ('Groups proteins from mzid2tsv output (single file '
                   'passed to -i)')

    def __init__(self):
        super().__init__()
        self.infiletype = 'TSV PSM table (MSGF+)'

    def create_lookup(self):
        lookups.build_proteingroup_db(self.lookup)



from app.lookups.sqlite.protpeptable import ProtPepTable


class ProtTableDB(ProtPepTable):
    datatype = 'protein'
    colmap = {'protein_group_master': ['master_id', 'protein_acc'],
              'proteins': ['pacc_id', 'protein_acc'],
              'protein_precur_quanted': ['pacc_id', 'prottable_id', 'quant'],
              'protein_fdr': ['pacc_id', 'prottable_id', 'fdr'],
              'protein_pep': ['pacc_id', 'prottable_id', 'pep'],
              'protein_probability': ['pacc_id', 'prottable_id',
                                      'probability'],
              'protquant_channels': ['channel_id', 'prottable_id',
                                     'channel_name', 'amount_psms_name'],
              'protein_iso_quanted': ['proteinquant_id', 'pacc_id',
                                      'channel_id', 'quantvalue',
                                      'amount_psms'],
              }

    def add_tables(self):
        self.create_tables(['protein_tables', 'protein_iso_quanted',
                            'protquant_channels', 'protein_precur_quanted',
                            'protein_probability', 'protein_fdr',
                            'protein_pep'])

    def get_unique_peptide_nrs(self):
        return self.get_unique_peptide_nrs_base('protein_acc', 'JOIN protein_group_content USING(protein_acc) JOIN protein_group_master AS acctable USING(master_id)')

    def get_precursorquant_headerfields(self):
        cursor = self.get_cursor()
        cursor.execute(
            'SELECT DISTINCT prottable_id '
            'FROM protein_precur_quanted')
        return cursor


class GeneTableDB(ProtPepTable):
    datatype = 'gene'
    colmap = {'genes': ['gene_id', 'gene_acc', 'protein_acc'],
              'gene_precur_quanted': ['gene_id', 'genetable_id', 'quant'],
              'gene_fdr': ['gene_id', 'genetable_id', 'fdr'],
              'gene_pep': ['gene_id', 'genetable_id', 'pep'],
              'gene_probability': ['gene_id', 'genetable_id',
                                   'probability'],
              'genequant_channels': ['channel_id', 'genetable_id',
                                     'channel_name', 'amount_psms_name'],
              'gene_iso_quanted': ['genequant_id', 'gene_id',
                                   'channel_id', 'quantvalue', 'amount_psms'],
              }

    def add_tables(self):
        self.create_tables(['gene_tables', 'gene_iso_quanted',
                            'genequant_channels', 'gene_precur_quanted',
                            'gene_probability', 'gene_fdr',
                            'gene_pep'])

    def get_protein_gene_symbol_for_map(self):
        fields = ['p.gene_acc', 'pd.description',
                  'aid.assoc_id', 'prot.protein_acc']
        genetable = self.table_map[self.datatype]['feattable']
        sql = (
                'SELECT {} FROM {} AS p '
                'LEFT OUTER JOIN prot_desc AS pd USING(protein_acc) '
                'LEFT OUTER JOIN associated_ids AS aid USING(protein_acc) '
                'LEFT OUTER JOIN proteins AS prot USING(protein_acc)')
        cursor = self.get_cursor()
        sql = sql.format(','.join(fields), genetable)
        return cursor.execute(sql)

    def get_proteins_psms_for_map(self):
        """Gets gene-PSM combinations from DB and filters out uniques
        on the fly. Filtering is done since PSM are stored per protein,
        not per gene, so there may be a lot of *plicates"""
        fields = ['p.gene_acc', 'sets.set_name',
                  'pep.sequence', 'psm.psm_id']
        firstjoin = ('protein_psm', 'pp', 'protein_acc')
        genetable = self.table_map[self.datatype]['feattable']
        return self.get_unique_gene_psms(genetable, fields, firstjoin)
             
    def get_unique_peptide_nrs(self):
        return self.get_unique_peptide_nrs_base('gene_acc', 'JOIN genes AS acctable USING(protein_acc)')

    def get_precursorquant_headerfields(self):
        cursor = self.get_cursor()
        cursor.execute(
            'SELECT DISTINCT prottable_id '
            'FROM {}'.format(self.table_map[self.datatype]['prectable']))
        return cursor


class GeneTableAssocIDsDB(GeneTableDB):
    datatype = 'assoc'

    def __init__(self, fn=None):
        super().__init__(fn)
        self.colmap.pop('genes')
        self.colmap = {table.replace('gene', 'assoc'): cols
                       for table, cols in self.colmap.items()}
        self.colmap['genequant_channels'] = self.colmap.pop(
            'assocquant_channels')
        self.colmap['associated_ids'] = ['gene_id', 'assoc_id', 'protein_acc']

    def add_tables(self):
        self.create_tables(['gene_tables', 'assoc_iso_quanted',
                            'genequant_channels', 'assoc_precur_quanted',
                            'assoc_probability', 'assoc_fdr',
                            'assoc_pep'])

    def get_unique_peptide_nrs(self):
        return self.get_unique_peptide_nrs_base('assoc_id', 'JOIN associated_ids AS acctable USING(protein_acc)')

    def get_protein_gene_symbol_for_map(self):
        fields = ['p.assoc_id', 'pd.description',
                  'gene.gene_acc', 'prot.protein_acc']
        genetable = self.table_map[self.datatype]['feattable']
        sql = (
                'SELECT {} FROM {} AS p '
                'LEFT OUTER JOIN prot_desc AS pd USING(protein_acc) '
                'LEFT OUTER JOIN genes AS gene USING(protein_acc) '
                'LEFT OUTER JOIN proteins AS prot USING(protein_acc)')
        cursor = self.get_cursor()
        sql = sql.format(','.join(fields), genetable)
        return cursor.execute(sql)

    def get_proteins_psms_for_map(self):
        """Gets gene-PSM combinations from DB and filters out uniques
        on the fly. Filtering is done since PSM are stored per protein,
        not per gene, so there may be a lot of *plicates"""
        fields = ['p.assoc_id', 'sets.set_name',
                  'pep.sequence', 'psm.psm_id']
        firstjoin = ('protein_psm', 'pp', 'protein_acc')
        genetable = self.table_map[self.datatype]['feattable']
        return self.get_unique_gene_psms(genetable, fields, firstjoin)



import os

from app.lookups import base as lookups
from app.drivers.base import BaseDriver
from app.drivers.options import mslookup_options


class LookupDriver(BaseDriver):
    outfile, outdir = None, None

    def __init__(self):
        super().__init__()
        self.parser_options = mslookup_options

    def set_options(self):
        super().set_options()
        del(self.options['-o'])
        del(self.options['-d'])
        self.options.update(self.define_options(['lookupfn'],
                                                mslookup_options))

    def initialize_lookup(self):
        if self.lookup is None:
            # FIXME MUST be a set or mzml lookup? here is place to assert
            # correct lookuptype!
            if self.outfile is None:
                self.outfile = os.path.join(self.outdir,
                                            'mslookup_db.sqlite')
            self.lookup = lookups.create_new_lookup(self.outfile,
                                                    self.lookuptype)
        self.lookup.add_tables()

    def run(self):
        self.initialize_lookup()
        self.create_lookup()



from app.drivers.prottable.base import ProttableAddData
from app.readers import tsv as tsvreader
from app.actions.prottable import bestpeptide as prep
from app.drivers.options import prottable_options


class BestPeptidePerProtein(ProttableAddData):
    """Filters peptide table to extract the best scoring peptides per protein,
    then adds this score to the protein table.
    """
    outsuffix = '_bestpep.tsv'
    command = 'bestpeptide'
    commandhelp = ('Given the protein table and corresponding peptide table '
                   ', fetch the best scoring peptide for each protein '
                   'and annotate that score in the protein table.')

    def set_options(self):
        super().set_options()
        self.options.update(self.define_options(['pepfile', 'proteincol',
                                                 'scorecolpattern',
                                                 'pcolpattern', 'minlogscore'],
                                                prottable_options))

    def parse_input(self, **kwargs):
        super().parse_input(**kwargs)
        self.headertypes = ['bestpepscore']

    def initialize_input(self):
        super().initialize_input()
        self.pepheader = tsvreader.get_tsv_header(self.pepfile)
        if self.proteincol:
            self.get_column_header_for_number(['proteincol'], self.pepheader)
        elif self.pcolpattern:
            self.proteincol = tsvreader.get_cols_in_file(self.pcolpattern,
                                                         self.pepheader, True)
        self.scorecol = tsvreader.get_cols_in_file(self.scorecolpattern,
                                                   self.pepheader, True)

    def set_feature_generator(self):
        self.features = prep.generate_proteins(self.pepfile, self.in_proteins,
                                               self.pepheader, self.scorecol,
                                               self.minlogscore,
                                               protcol=self.proteincol)



import os

from app.actions.mzidtsv import splitmerge as prep
from app.drivers.mzidtsv import MzidTSVDriver
from app.writers import mzidtsv as writers
from app.drivers.options import mzidtsv_options


class MzidTSVConcatenateDriver(MzidTSVDriver):
    """Concatenates TSVs"""
    outsuffix = '_concat.tsv'
    command = 'merge'
    commandhelp = ('Merges multiple TSV tables of MSGF+ output.'
                   'Make sure headers are same in all files.')

    def set_options(self):
        super().set_options()
        self.options.update(self.define_options(['multifiles'],
                                                mzidtsv_options))

    def get_psms(self):
        self.header = self.oldheader
        self.psms = prep.merge_mzidtsvs(self.fn, self.oldheader)


class MzidTSVSplitDriver(MzidTSVDriver):
    """Splits MSGF PSM table on contents of certain column. Each
    row in file is piped to an output file. Which output files
    the row is written to depends on the contents of the selected
    column"""
    outsuffix = '_split.tsv'
    command = 'split'
    commandhelp = 'Splits an MSGF TSV PSM table into multiple new tables'

    def set_options(self):
        super().set_options()
        options = self.define_options(['bioset', 'splitcol'],
                                      mzidtsv_options)
        self.options.update(options)

    def get_psms(self):
        self.header = self.oldheader[:]
        self.psms = prep.generate_psms_split(self.fn, self.oldheader,
                                             self.bioset, self.splitcol)

    def write(self):
        base_outfile = os.path.join(self.outdir, '{}.tsv')
        writers.write_multi_mzidtsv(self.header, self.oldheader, self.psms,
                                    base_outfile)



import sys
from statistics import median, StatisticsError
from collections import OrderedDict

from app.dataformats import prottable as prottabledata
from app.readers import tsv as reader
from app.actions import isonormalizing


ISOQUANTRATIO_FEAT_ACC = '##isoquant_target_acc##'


def get_isobaric_ratios(psmfn, psmheader, channels, denom_channels, min_int,
                        targetfn, accessioncol, normalize, normratiofn):
    """Main function to calculate ratios for PSMs, peptides, proteins, genes.
    Can do simple ratios, median-of-ratios and median-centering
    normalization."""
    psm_or_feat_ratios = get_psmratios(psmfn, psmheader, channels,
                                       denom_channels, min_int, accessioncol)
    if normalize and normratiofn:
        normheader = reader.get_tsv_header(normratiofn)
        normratios = get_ratios_from_fn(normratiofn, normheader, channels)
        ch_medians = get_medians(channels, normratios, report=True)
        outratios = calculate_normalized_ratios(psm_or_feat_ratios, ch_medians,
                                                channels)
    elif normalize:
        flatratios = [[feat[ch] for ch in channels]
                      for feat in psm_or_feat_ratios]
        ch_medians = get_medians(channels, flatratios, report=True)
        outratios = calculate_normalized_ratios(psm_or_feat_ratios, ch_medians,
                                                channels)
    else:
        outratios = psm_or_feat_ratios
    # at this point, outratios look like:
    # [{ch1: 123, ch2: 456, ISOQUANTRATIO_FEAT_ACC: ENSG1244}, ]
    if accessioncol and targetfn:
        outratios = {x[ISOQUANTRATIO_FEAT_ACC]: x for x in outratios}
        return output_to_target_accession_table(targetfn, outratios, channels)
    elif not accessioncol and not targetfn:
        return paste_to_psmtable(psmfn, psmheader, outratios)
    elif accessioncol and not targetfn:
        # generate new table with accessions
        return ({(k if not k == ISOQUANTRATIO_FEAT_ACC
                  else prottabledata.HEADER_ACCESSION): v
                 for k, v in ratio.items()} for ratio in outratios)


def get_psmratios(psmfn, header, channels, denom_channels, min_int, acc_col):
    allfeats, feat_order, psmratios = {}, OrderedDict(), []
    for psm in reader.generate_tsv_psms(psmfn, header):
        ratios = calc_psm_ratios(psm, channels, denom_channels, min_int)
        # remove uninformative psms when adding to features
        if acc_col and (psm[acc_col] == '' or ';' in psm[acc_col] or
                        not {psm[q] for q in channels}.difference(
                            {'NA', None, False, ''})):
            continue
        elif acc_col:
            try:
                allfeats[psm[acc_col]].append(ratios)
            except KeyError:
                allfeats[psm[acc_col]] = [ratios]
            feat_order[psm[acc_col]] = 1
        else:
            psmquant = {ch: str(ratios[ix]) if ratios[ix] != 'NA' else 'NA'
                        for ix, ch in enumerate(channels)}
            psmquant[ISOQUANTRATIO_FEAT_ACC] = False
            psmratios.append(psmquant)
    if not acc_col:
        return psmratios
    else:
        outfeatures = []
        for feat in feat_order.keys():
            quants = allfeats[feat]
            outfeature = {ISOQUANTRATIO_FEAT_ACC: feat}
            outfeature.update(get_medians(channels, quants))
            outfeature.update(get_no_psms(channels, quants))
            outfeatures.append(outfeature)
    return outfeatures


def get_ratios_from_fn(fn, header, channels):
    ratios = []
    for feat in reader.generate_tsv_psms(fn, header):
        ratios.append([float(feat[ch]) if feat[ch] != 'NA' else 'NA'
                       for ch in channels])
    return ratios


def paste_to_psmtable(psmfn, header, ratios):
    # loop psms in psmtable, paste the outratios in memory
    for psm, ratio in zip(reader.generate_tsv_psms(psmfn, header), ratios):
        ratio.pop(ISOQUANTRATIO_FEAT_ACC)
        ratio = {'ratio_{}'.format(ch): val for ch, val in ratio.items()}
        psm.update(ratio)
        yield psm


def output_to_target_accession_table(targetfn, featratios, channels):
    #loop prottable, add ratios from dict, acc = key
    theader = reader.get_tsv_header(targetfn)
    acc_field = theader[0]
    for feat in reader.generate_tsv_proteins(targetfn, theader):
        try:
            quants = featratios[feat[acc_field]]
        except KeyError:
            quants = {ch: 'NA' for ch in channels}
            quants.update({get_no_psms_field(ch): 'NA' for ch in channels})
        else:
            quants.pop(ISOQUANTRATIO_FEAT_ACC)
        feat.update(quants)
        yield feat


def calc_psm_ratios(psm, channels, denom_channels, min_intensity):
    # set values below min_intensity to NA
    psm_intensity = {ch: float(psm[ch])
                     if psm[ch] != 'NA' and float(psm[ch]) > min_intensity
                     else 'NA' for ch in channels}
    denomvalues = [psm_intensity[ch] for ch in denom_channels
                   if psm_intensity[ch] != 'NA']
    if sum(denomvalues) == 0 or len(denomvalues) == 0:
        return ['NA'] * len(channels)
    denom = sum(denomvalues) / len(denomvalues)
    return [psm_intensity[ch] / denom
            if psm_intensity[ch] != 'NA' else 'NA' for ch in channels]


def get_medians(channels, ratios, report=False):
    ch_medians = {}
    for ix, channel in enumerate(channels):
        try:
            ch_medians[channel] = median([x[ix] for x in ratios
                                          if x[ix] != 'NA'])
        except StatisticsError:
            # channel is empty, common in protein quant but not in normalizing
            ch_medians[channel] = 'NA'
    if report:
        report = ('Channel intensity medians used for normalization:\n'
                  '{}'.format('\n'.join(['{} - {}'.format(ch, ch_medians[ch])
                                         for ch in channels])))
        sys.stdout.write(report)
    return ch_medians


def get_no_psms_field(quantfield):
    return '{}{}'.format(quantfield, prottabledata.HEADER_NO_PSMS_SUFFIX)


def get_no_psms(channels, ratios):
    ch_nopsms = {}
    for ix, channel in enumerate(channels):
        fieldname = get_no_psms_field(channel)
        ch_nopsms[fieldname] = len([x[ix] for x in ratios if x[ix] != 'NA'])
    return ch_nopsms


def calculate_normalized_ratios(ratios, ch_medians, channels):
    """Calculates ratios for PSM tables containing isobaric channels with
    raw intensities. Normalizes the ratios by median. NA values or values
    below min_intensity are excluded from the normalization."""
    outratios = []
    for quant in ratios:
        quant.update({ch: str(quant[ch] / ch_medians[ch])
                      if quant[ch] != 'NA' else 'NA' for ch in channels})
        outratios.append(quant)
    return outratios


def get_normalized_ratios(psmfn, header, channels, denom_channels,
                          min_intensity, second_psmfn, secondheader):
    """Calculates ratios for PSM tables containing isobaric channels with
    raw intensities. Normalizes the ratios by median. NA values or values
    below min_intensity are excluded from the normalization."""
    ratios = []
    if second_psmfn is not None:
        median_psmfn = second_psmfn
        medianheader = secondheader
    else:
        median_psmfn = psmfn
        medianheader = header
    for psm in reader.generate_tsv_psms(median_psmfn, medianheader):
        ratios.append(calc_psm_ratios(psm, channels, denom_channels,
                                      min_intensity))
    ch_medians = isonormalizing.get_medians(channels, ratios)
    report = ('Channel intensity medians used for normalization:\n'
              '{}'.format('\n'.join(['{} - {}'.format(ch, ch_medians[ch])
                                     for ch in channels])))
    sys.stdout.write(report)
    for psm in reader.generate_tsv_psms(psmfn, header):
        psmratios = calc_psm_ratios(psm, channels, denom_channels,
                                    min_intensity)
        psm.update({ch: str(psmratios[ix] / ch_medians[ch])
                    if psmratios[ix] != 'NA' else 'NA'
                    for ix, ch in enumerate(channels)})
        yield psm



import os

from app.actions.prottable import merge as preparation
from app.drivers.prottable.base import ProttableMergeDriver
from app.drivers.options import prottable_options


class BuildProteinTableDriver(ProttableMergeDriver):
    outsuffix = ''
    lookuptype = 'prottable'
    command = 'build'
    commandhelp = ('Create protein table from a lookup database. '
                   'E.g. when multiple protein quant tables have '
                   'been read into the lookup and will be combined.'
                   'Does not need an -i.')

    def set_options(self):
        super().set_options()
        self.options.update(self.define_options(['lookupfn', 'genecentric',
                                                 'isobaric', 'precursor',
                                                 'probability', 'fdr', 'pep',
                                                 'mock_infn', 'mergecutoff'],
                                                prottable_options))

    def parse_input(self, **kwargs):
        """Build protein table has no input file (though it has a lookup),
        which is why we set it to outfile name so the infile fetching
        and outfile creating wont error."""
        super().parse_input(**kwargs)
        self.fn = os.path.join(os.getcwd(), 'built_protein_table.txt')
        if self.genecentric:
            self.lookuptype = {'genes': 'genetable',
                               'assoc': 'associdtable'}[self.genecentric]

    def set_feature_generator(self):
        """Generates proteins with quant from the lookup table"""
        self.features = preparation.build_proteintable(self.lookup,
                                                       self.headerfields,
                                                       self.mergecutoff,
                                                       self.isobaric,
                                                       self.precursor,
                                                       self.probability,
                                                       self.fdr, self.pep,
                                                       self.genecentric)



def add_protgene_to_protdata(proteindata, p_acc, pgenedata, genecentric, pgcontentmap):
    if genecentric:
        desc = pgenedata[1]
        cov, pgcontent = None, None
        if genecentric == 'genes':
            gene = None
            assoc_id, protid = pgenedata[2], pgenedata[3]
        else:
            assoc_id=None
            gene, protid = pgenedata[2], pgenedata[3]
    else:
        desc, cov = pgenedata[2], pgenedata[3]
        gene, assoc_id = pgenedata[4], pgenedata[5]
        pgcontent = pgcontentmap[p_acc]
        protid = None
    pdata = {'desc': desc, 'pgcontent': pgcontent, 'cov': cov}
    try:
        proteindata[p_acc].update(pdata)
    except KeyError:
        proteindata[p_acc] = pdata
        proteindata[p_acc].update({'gene': set(), 'aid': set(), 'protein_ids': set()})
    proteindata[p_acc]['gene'].add(gene)
    proteindata[p_acc]['aid'].add(assoc_id)
    proteindata[p_acc]['protein_ids'].add(protid)


def add_psms_to_proteindata(proteindata, p_acc, pool, psmdata):
    """Fill function for create_featuredata_map"""
    seq, psm_id = psmdata[2], psmdata[3]
    try:
        proteindata[p_acc]['pools'][pool]['psms'].add(psm_id)
    except KeyError:
        emptyinfo = {'psms': set(), 'peptides': set(), 'unipeps': 0}
        try:
            proteindata[p_acc]['pools'][pool] = emptyinfo
        except KeyError:
            proteindata[p_acc].update({'pools': {pool: emptyinfo}})
        proteindata[p_acc]['pools'][pool]['psms'].add(psm_id)
    proteindata[p_acc]['pools'][pool]['peptides'].add(seq)


def create_featuredata_map(pgdb, psm_fill_fun, pgene_fill_fun, genecentric=False, 
        count_fun=None, pool_to_output=False, get_uniques=False, is_peptides=False):
    """Creates dict of protein data containing PSMs, peptides, proteins in
    protein group, unique peptides, description and coverage. Loops through
    PSM/protein matches and uses a passed fill_fun function to actually
    fill the outputted map.
    """
    if genecentric:
        pgcontentmap = None
    else:
        pgcontentmap = get_proteingroup_content(pgdb)
    proteindata = {}
    for pgenedata in pgdb.get_protein_gene_symbol_for_map():
        if is_peptides:
            # include sequence (index 1 of record) to fill func
            pgene_fill_fun(proteindata, pgenedata[1], pgenedata[0], 
                    pgenedata, genecentric, pgcontentmap)
        else:
            pgene_fill_fun(proteindata, pgenedata[0], 
                    pgenedata, genecentric, pgcontentmap)
    protein_psms_data = pgdb.get_proteins_psms_for_map()
    psmdata = next(protein_psms_data)
    last_prot, last_pool = psmdata[0], psmdata[1]
    psm_fill_fun(proteindata, last_prot, last_pool, psmdata)
    for psmdata in protein_psms_data:
        p_acc, samplepool = psmdata[0], psmdata[1]
        if pool_to_output and samplepool != pool_to_output:
            continue
        if samplepool != last_pool or p_acc != last_prot:
            if count_fun is not None:
                count_fun(proteindata, last_prot, last_pool)
            last_pool, last_prot = samplepool, p_acc
        psm_fill_fun(proteindata, p_acc, samplepool, psmdata)
    if count_fun is not None:
        count_fun(proteindata, last_prot, last_pool)
    
    if get_uniques:
        for no_uni, pool, pacc in pgdb.get_unique_peptide_nrs():
            proteindata[pacc]['pools'][pool]['unipeps'] = no_uni
    return proteindata


def get_proteingroup_content(pgdb):
    pgmap = {}
    for master, contentprot in pgdb.get_proteingroup_content():
        try:
            pgmap[master].append(contentprot)
        except KeyError:
            pgmap[master] = [contentprot]
    return pgmap



from app.drivers.prottable.base import PepProttableDriver
from app.actions.headers import peptable as head
from app.readers import tsv as tsvreader
from app.actions.peptable import psmtopeptable as prep
from app.drivers.options import peptable_options


class MzidTSVPeptableDriver(PepProttableDriver):
    """Creates unique peptide table from MzidTSV table. Will not change
    q-values/PEP after filtration, so these should have been calculated
    for the peptides beforehand or rescored afterwards.
    """
    outsuffix = '_peptable.tsv'
    command = 'psm2pep'
    commandhelp = ('Create peptide table from PSM TSV input, uses best '
                   'scoring PSM for each peptide and strips PSM '
                   'isobaric quant information. Retains MS1 quant info if '
                   'specified, by taking the highest precursor quant value '
                   'for a peptide.')

    def __init__(self):
        super().__init__()
        self.infiletype = 'TSV PSM table (MSGF+)'

    def set_options(self):
        super().set_options()
        self.options.update(self.define_options(['spectracol',
                                                 'scorecolpattern',
                                                 'quantcolpattern',
                                                 'precursorquantcolpattern'],
                                                peptable_options))
        self.options['--isobquantcolpattern']['help'] += (
            'Isobaric quantification values will be removed from the table '
            'since they represent PSMs')

    def initialize_input(self):
        self.oldheader = tsvreader.get_tsv_header(self.fn)
        self.get_column_header_for_number(['spectracol'])
        self.scorecol = tsvreader.get_cols_in_file(self.scorecolpattern,
                                                   self.oldheader, True)
        self.precurquantcol = prep.get_quantcols(self.precursorquantcolpattern,
                                                 self.oldheader, 'precur')

    def create_header(self):
        self.header = head.get_psm2pep_header(self.oldheader,
                                              self.quantcolpattern,
                                              self.precurquantcol)

    def set_feature_generator(self):
        self.features = prep.generate_peptides(self.fn, self.oldheader,
                                               self.scorecol,
                                               self.precurquantcol,
                                               self.spectracol)



from app.actions.mzidtsv import percolator as prep
from app.drivers.mzidtsv import MzidTSVDriver
from app.drivers.options import mzidtsv_options


class MzidPercoTSVDriver(MzidTSVDriver):
    """
    Adds percolator data from mzid file to table.
    """
    outsuffix = '_percolated.tsv'
    command = 'percolator'
    commandhelp = ('Add percolator data to a  TSV with MSGF+ output. ')

    def set_options(self):
        super().set_options()
        self.options.update(self.define_options(['mzidfn'], mzidtsv_options))

    def get_psms(self):
        self.multipsm_per_scan = False
        if self.multipsm_per_scan:
            # FIXME not supported yet
            # Create mzid PSM/sequence sqlite (fn, scan, rank, sequence)
            pass
        self.header = prep.get_header_with_percolator(self.oldheader,
                                                      self.multipsm_per_scan)
        self.psms = prep.add_percolator_to_mzidtsv(self.mzidfn,
                                                   self.fn,
                                                   self.multipsm_per_scan,
                                                   self.oldheader)



from app.drivers.mslookup import base
from app.readers import spectra as spectrareader
from app.actions.mslookup import spectra as spectralookup
from app.actions.mslookup import biosets as biosetlookup
from app.drivers.options import mslookup_options


class SpectraLookupDriver(base.LookupDriver):
    lookuptype = 'spectra'
    command = 'spectra'
    commandhelp = ('Create lookup of spectra in mzML '
                   'format. Requires passing mzML files to -i, but '
                   'neither --spectra nor --dbfile. Biological set names '
                   'for each file should be specified using --setnames')

    def __init__(self):
        super().__init__()
        self.infiletype = 'mzML spectra'

    def parse_input(self, **kwargs):
        super().parse_input(**kwargs)
        self.spectrafns = self.fn
        if self.setnames is None:
            assert self.lookup is not None, ('Must specify lookup '
                                             'if setnames have not '
                                             'been provided')
        else:
            self.setnames = [x.replace('"', '') for x in self.setnames]

    def set_options(self):
        super().set_options()
        self.options['--dbfile'].update({'required': False, 'default': None})
        self.options.update(self.define_options(['multifiles', 'setnames'],
                                                mslookup_options))

    def create_lookup(self):
        biosetlookup.create_bioset_lookup(self.lookup, self.spectrafns,
                                          self.setnames)
        fn_spectra = spectrareader.mzmlfn_ms2_spectra_generator(
            self.spectrafns)
        spectralookup.create_spectra_lookup(self.lookup, fn_spectra)



from Bio import SeqIO
PROTEIN_STORE_CHUNK_SIZE = 100000


def create_searchspace_wholeproteins(lookup, fastafn, minpeplen):
    fasta = SeqIO.parse(fastafn, 'fasta')
    prots = {str(prot.seq).replace('L', 'I'): prot.id for prot in fasta}
    storeseqs = []
    peptotal = 0
    for protseq, prot_id in prots.items():
        for pos in range(0, len(protseq) - minpeplen + 1):
            possible_pep = protseq[pos:pos + minpeplen]
            peptotal += 1
            storeseqs.append((possible_pep, prot_id, pos))
        if len(storeseqs) > PROTEIN_STORE_CHUNK_SIZE:
            lookup.store_pep_proteins(storeseqs)
            storeseqs = []
    lookup.store_pep_proteins(storeseqs)
    lookup.index_proteins()
    print('Stored {} peptides from {} proteins (reduced FASTA to remove '
          'duplicate sequences)'.format(peptotal, len(prots)))


def create_searchspace(lookup, fastafn, proline_cut=False,
                       reverse_seqs=True, do_trypsinize=True):
    """Given a FASTA database, proteins are trypsinized and resulting peptides
    stored in a database or dict for lookups"""
    allpeps = []
    for record in SeqIO.parse(fastafn, 'fasta'):
        if do_trypsinize:
            pepseqs = trypsinize(record.seq, proline_cut)
        else:
            pepseqs = [record.seq]
        # Exchange all leucines to isoleucines because MS can't differ
        pepseqs = [(str(pep).replace('L', 'I'),) for pep in pepseqs]
        allpeps.extend(pepseqs)
        if len(allpeps) > 1000000:  # more than x peps, write to SQLite
            lookup.write_peps(allpeps, reverse_seqs)
            allpeps = []
    # write remaining peps to sqlite
    lookup.write_peps(allpeps, reverse_seqs)
    lookup.index_peps(reverse_seqs)
    lookup.close_connection()


def trypsinize(proseq, proline_cut=False):
    # TODO add cysteine to non cut options, use enums
    """Trypsinize a protein sequence. Returns a list of peptides.
    Peptides include both cut and non-cut when P is behind a tryptic
    residue. Multiple consequent tryptic residues are treated as follows:
    PEPKKKTIDE - [PEPK, PEPKK, PEPKKK, KKTIDE, KTIDE, TIDE, K, K, KK ]
    """
    outpeps = []
    currentpeps = ['']
    trypres = set(['K', 'R'])
    noncutters = set()
    if not proline_cut:
        noncutters.add('P')
    for i, aa in enumerate(proseq):
        currentpeps = ['{0}{1}'.format(x, aa) for x in currentpeps]
        if i == len(proseq) - 1:
            continue
        if aa in trypres and proseq[i + 1] not in noncutters:
            outpeps.extend(currentpeps)  # do actual cut by storing peptides
            if proseq[i + 1] in trypres.union('P'):
                # add new peptide to list if we are also to run on
                currentpeps.append('')
            elif trypres.issuperset(currentpeps[-1]):
                currentpeps = [x for x in currentpeps if trypres.issuperset(x)]
                currentpeps.append('')
            else:
                currentpeps = ['']

    if currentpeps != ['']:
        outpeps.extend(currentpeps)
    return outpeps



from app.dataformats import prottable as prottabledata
from app.actions.shared.pepprot_isoquant import base_add_isoquant_data


def get_isoquant_header(quantfields):
    baseheader = [prottabledata.HEADER_PROTEIN] + quantfields
    nopsms = [get_no_psms_field(qf) for qf in quantfields]
    return baseheader + nopsms


def get_no_psms_field(quantfield):
    return '{}{}'.format(quantfield, prottabledata.HEADER_NO_PSMS_SUFFIX)


def add_isoquant_data(proteins, quantproteins, quantacc, quantfields):
    """Runs through a protein table and adds quant data from ANOTHER protein
    table that contains that data."""
    for protein in base_add_isoquant_data(proteins, quantproteins,
                                          prottabledata.HEADER_PROTEIN,
                                          quantacc, quantfields):
        yield protein


def get_no_psms(channels, ratios):
    ch_nopsms = {}
    for ix, channel in enumerate(channels):
        fieldname = get_no_psms_field(channel)
        ch_nopsms[fieldname] = len([x[ix] for x in ratios if x[ix] != 'NA'])
    return ch_nopsms



from app.readers import tsv as readers
from app.dataformats import mzidtsv as mzidtsvdata


def create_header(oldheader, spectracol):
    newheadings = mzidtsvdata.SPECDATA_HEADER
    header = oldheader[:spectracol] + newheadings + oldheader[spectracol:]
    return header


def generate_psms_spectradata(lookup, tsvfn, oldheader):
    psm_specdata = zip(enumerate(readers.generate_tsv_psms(tsvfn, oldheader)),
                       lookup.get_exp_spectra_data_rows())
    for (row, psm), specdata in psm_specdata:
        outpsm = {x: y for x, y in psm.items()}
        if row == int(specdata[0]):
            outpsm.update({mzidtsvdata.HEADER_SETNAME: specdata[1],
                           mzidtsvdata.HEADER_RETENTION_TIME: str(specdata[2]),
                           mzidtsvdata.HEADER_INJECTION_TIME: str(specdata[3]),
                           })
        else:
            raise RuntimeError('PSM with row nr {} has no rownr in DB. '
                               'Current DB row is {}'.format(row, specdata[0]))
        yield outpsm



from statistics import median

from app.dataformats import mzidtsv as mzidtsvdata
from app.dataformats import peptable as peptabledata
from app.readers import tsv as reader
from app.actions.peptable.base import evaluate_peptide
from app.actions.headers.peptable import switch_psm_to_peptable_fields


def get_quantcols(pattern, oldheader, coltype):
    """Searches for quantification columns using pattern and header list.
    Calls reader function to do regexp. Returns a single column for
    precursor quant."""
    if pattern is None:
       return False
    if coltype == 'precur':
        return reader.get_cols_in_file(pattern, oldheader, single_col=True)


def generate_peptides(tsvfn, oldheader, scorecol, precurquantcol, fncol=None,
                      higherbetter=True):
    if fncol is None:
        fncol = mzidtsvdata.HEADER_SPECFILE
    peptides = {}
    switch_map = switch_psm_to_peptable_fields(oldheader)
    for psm in reader.generate_tsv_psms(tsvfn, oldheader):
        for oldkey, newkey in switch_map.items():
            try:
                psm[newkey] = psm.pop(oldkey)
            except KeyError:
                pass
        pepseq = psm[peptabledata.HEADER_PEPTIDE]
        peptides = evaluate_peptide(peptides, psm, pepseq, higherbetter,
                                    scorecol, fncol)
        add_quant_values(peptides, psm, precurquantcol)
    for peptide in peptides.values():
        peptide['line'][peptabledata.HEADER_LINKED_PSMS] = '; '.join(
            peptide['psms'])
        for qtype, pepquant in peptide['quant'].items():
            peptide['line'].update(parse_quant_data(qtype, pepquant))
        yield peptide['line']


def parse_quant_data(qtype, pepquant):
    if qtype == 'precur':
        quants = {peptabledata.HEADER_AREA: get_peptide_quant(pepquant, 'precur')}
    return quants


def get_peptide_quant(quantdata, quanttype):
    """Parses lists of quantdata and returns maxvalue from them. Strips NA"""
    parsefnx = {'precur': max}
    quantfloats = []
    for q in quantdata:
        try:
            quantfloats.append(float(q))
        except(TypeError, ValueError):
            pass
    if not quantfloats:
        return 'NA'
    return str(parsefnx[quanttype](quantfloats))


def add_quant_values(allpeps, psm, precurq_field):
    try:
        quants = allpeps[psm[peptabledata.HEADER_PEPTIDE]]['quant']
    except KeyError:
        quants = {}
    if precurq_field:
        try:
            quants['precur'].append(psm[precurq_field])
        except KeyError:
            quants['precur'] = [psm[precurq_field]]
    try:
        allpeps[psm[peptabledata.HEADER_PEPTIDE]]['quant'] = quants
    except KeyError:
        # PSM sequence does not exist in allpeps because has no good-scoring peptides
        pass



def passes_filter(psm, threshold, confkey, lower_is_better):
    if psm[confkey] in ['NA', '', None, False]:
        return False
    lower = float(psm[confkey]) < float(threshold)
    return lower == lower_is_better



import os
from app.readers import tsv as tsvreader


def create_peptidequant_lookup(fns, pqdb, poolnames, pepseq_colnr,
                               ms1_qcolpattern=None, isobqcolpattern=None,
                               psmnrpattern=None, fdrcolpattern=None,
                               pepcolpattern=None):
    """Calls lower level function to create a peptide quant lookup"""
    patterns = [ms1_qcolpattern, fdrcolpattern, pepcolpattern]
    storefuns = [pqdb.store_precursor_quants, pqdb.store_fdr,
                 pqdb.store_pep]
    create_pep_protein_quant_lookup(fns, pqdb, poolnames, pepseq_colnr,
                                    patterns, storefuns,
                                    isobqcolpattern, psmnrpattern)


def create_proteinquant_lookup(fns, pqdb, poolnames, protacc_colnr,
                               ms1_qcolpattern=None, isobqcolpattern=None,
                               psmnrpattern=None, probcolpattern=None,
                               fdrcolpattern=None, pepcolpattern=None):
    """Calls lower level function to create a protein quant lookup"""
    patterns = [ms1_qcolpattern, probcolpattern, fdrcolpattern, pepcolpattern]
    storefuns = [pqdb.store_precursor_quants, pqdb.store_probability,
                 pqdb.store_fdr, pqdb.store_pep]
    create_pep_protein_quant_lookup(fns, pqdb, poolnames, protacc_colnr,
                                    patterns, storefuns, isobqcolpattern,
                                    psmnrpattern)


def create_pep_protein_quant_lookup(fns, pqdb, poolnames, featcolnr, patterns,
                                    storefuns, isobqcolpattern=None,
                                    psmnrpattern=None):
    """Does the work when creating peptide and protein quant lookups. This
    loops through storing options and parses columns, passing on to the
    storing functions"""
    tablefn_map = create_tablefn_map(fns, pqdb, poolnames)
    feat_map = pqdb.get_feature_map()
    for pattern, storefun in zip(patterns, storefuns):
        if pattern is None:
            continue
        colmap = get_colmap(fns, pattern, single_col=True)
        if colmap:
            store_single_col_data(fns, tablefn_map, feat_map,
                                  storefun, featcolnr, colmap)
    if isobqcolpattern is not None:
        isocolmap = get_colmap(fns, isobqcolpattern, antipattern=psmnrpattern)
    else:
        return
    if psmnrpattern is not None:
        psmcolmap = get_colmap(fns, psmnrpattern)
    else:
        psmcolmap = False
    create_isobaric_quant_lookup(fns, tablefn_map,
                                 feat_map, pqdb,
                                 featcolnr,
                                 isocolmap, psmcolmap)


def create_tablefn_map(fns, pqdb, poolnames):
    """Stores protein/peptide table names in DB, returns a map with their
    respective DB IDs"""
    poolmap = {name: pid for (name, pid) in pqdb.get_all_poolnames()}
    pqdb.store_table_files([(poolmap[pool], os.path.basename(fn))
                            for fn, pool in zip(fns, poolnames)])
    return pqdb.get_tablefn_map()


def get_colmap(fns, pattern, single_col=False, antipattern=False):
    """For table files, loops through headers and checks which column(s)
    match a passed pattern. Those column(s) names are returned in a map with
    filenames as keys"""
    colmap = {}
    for fn in fns:
        header = tsvreader.get_tsv_header(fn)
        basefn = os.path.basename(fn)
        cols = tsvreader.get_cols_in_file(pattern, header, single_col)
        if antipattern:
            anticols = tsvreader.get_cols_in_file(antipattern, header,
                                                  single_col)
            cols = [col for col in cols if col not in anticols]
        if cols:
            colmap[basefn] = cols
    return colmap


def store_single_col_data(fns, prottable_id_map, pacc_map, pqdbmethod,
                          protacc_colnr, colmap):
    """General method to store single column data from protein tables
    in lookup"""
    to_store = []
    for fn, header, pquant in tsvreader.generate_tsv_pep_protein_quants(fns):
        pacc_id = pacc_map[pquant[header[protacc_colnr]]]
        pqdata = (pacc_id, prottable_id_map[fn], pquant[colmap[fn]])
        to_store.append(pqdata)
        if len(to_store) > 10000:
            pqdbmethod(to_store)
            to_store = []
    pqdbmethod(to_store)


def create_isobaric_quant_lookup(fns, tablefn_map, featmap, pqdb,
                                 featcolnr, allquantcols, psmcolmap):
    """Creates a lookup dict from peptide/protein quant input files and some
    input parameters. This assumes the order of quant columns and
    number-of-PSM columns is the same."""
    pqdb.store_quant_channels(
            map_psmnrcol_to_quantcol(allquantcols, psmcolmap, tablefn_map),
            psmcolmap)
    quantmap = pqdb.get_quantchannel_map()
    to_store = []
    for fn, header, pquant in tsvreader.generate_tsv_pep_protein_quants(fns):
        pqdata = get_isob_quant_data(pquant, header, tablefn_map[fn],
                                     featmap, featcolnr, quantmap)
        to_store.extend(pqdata)
        if len(to_store) > 10000:
            pqdb.store_isobaric_quants(to_store, psmcolmap)
            to_store = []
    pqdb.store_isobaric_quants(to_store, psmcolmap)


def map_psmnrcol_to_quantcol(quantcols, psmcols, tablefn_map):
    """This function yields tuples of table filename, isobaric quant column
    and if necessary number-of-PSM column"""
    if not psmcols:
        for fn in quantcols:
            for qcol in quantcols[fn]:
                yield (tablefn_map[fn], qcol)
    else:
        for fn in quantcols:
            for qcol, psmcol in zip(quantcols[fn], psmcols[fn]):
                yield (tablefn_map[fn], qcol, psmcol)


def get_isob_quant_data(pquant, header, fnid, featmap, featcol, qmap):
    """Turns a dict from a line of protein/peptide quant data into a set of
    tuples that can be stored in db"""
    feat_dbid = featmap[pquant[header[featcol]]]
    for channel_name, (channel_id, psmfield) in qmap[fnid].items():
        if psmfield is None:
            yield (feat_dbid, channel_id, pquant[channel_name])
        else:
            yield (feat_dbid, channel_id, pquant[channel_name],
                   pquant[psmfield])
